import os
import sys

from os.path import basename

from . import bin


def main():
    prog = basename(sys.argv[0])
    sys.argv = sys.argv[1:]

    arg_idx = None
    for idx, arg in enumerate(sys.argv):
        if arg != 'wara':
            arg_idx = idx
            break

    if arg_idx is None and len(sys.argv) > 2:
        print(
r'''
        ╭────────────────────────────────╮
        │       ╭──────────────────────╮ │
        │       │   ╭────────────────╮ │ │
        │       │   │ RecursionError │ │ │
        │       │   ╰──────┬─────────╯ │ │
        │       │     ,__, ╯           │ │
        │       │   __(..)             │ │
        │       │ ~(  (__)             │ │
        │       │  ||-||               │ │
        │       ╰────┬─────────────────╯ │
        │       ,__, ╯                   │
        │    ___(..)                     │
        │  /(   (__)                     │
        │ ' ||--||                       │
        ╰────────┬───────────────────────╯
            (__) ╯
    _______/(..)
  /(       /(__)
 * | w----||
   ||     ||
''', file=sys.stderr)
        sys.exit(1)

    sys.argv = sys.argv[(arg_idx or 0):]

    if not sys.argv:
        for f in sorted(os.listdir(os.path.dirname(__file__))):
            if f.startswith('bin_') and f.endswith('.py'):
                m = os.path.splitext(f[4:])[0]
                print(m)
        sys.exit(1)

    subcmd = sys.argv[0]

    try:
        getattr(bin, subcmd).main()
    except (AttributeError, ModuleNotFoundError):
        print(f'Unknown subcommand: {subcmd}', file=sys.stderr)
        sys.exit(1)



from .lib_test_utils import *

from warawara import *


class TestColorFacade(TestCase):
    def test_color_facade(self):
        # no argument unpack
        with self.raises(TypeError):
            color((208,))

        with self.raises(TypeError):
            color([208])

        with self.raises(TypeError):
            color((0xC0, 0xFF, 0xEE))

        with self.raises(TypeError):
            color([0xC0, 0xFF, 0xEE])

        with self.raises(TypeError):
            color(0xC0, 0xFF)

        # copy_ctor
        self.eq(color(color(208)), color(208))

        # subclass
        self.true(issubclass(Color256, Color))
        self.true(issubclass(ColorRGB, Color))

        # Color256
        orange = color(208)
        self.true(isinstance(orange, Color256))
        self.true(isinstance(orange, Color))

        # ColorRGB
        coffee = color(0xC0, 0xFF, 0xEE)
        self.true(isinstance(coffee, ColorRGB))
        self.true(isinstance(coffee, Color))

        # ColorRGB
        coffee = color('#C0FFEE')
        self.true(isinstance(coffee, ColorRGB))
        self.true(isinstance(coffee, Color))

        # ColorHSV
        lime = color('@120,100,100')
        self.true(isinstance(lime, ColorHSV))
        self.true(isinstance(lime, Color))

    def test_color_invalid_value(self):
        with self.raises(TypeError):
            color(True)

        with self.raises(TypeError):
            Color256(True)

        with self.raises(TypeError):
            ColorRGB(True)


class TestColorTraits(TestCase):
    def setUp(self):
        self.orange = color(214)
        self.coffee = color('#C0FFEE')
        self.purple = color('@300,100,50')

    def test_repr(self):
        self.eq(repr(self.orange), 'Color256(214)')
        self.eq(' '.join(repr(self.coffee).split()), 'ColorRGB(192, 255, 238)')
        self.eq(' '.join(repr(self.purple).split()), 'ColorHSV(300deg, 100%, 50%)')

    def test_int(self):
        self.eq(int(self.orange), 214)
        self.eq(int(self.coffee), 0xC0FFEE)
        self.eq(int(self.purple), 300100050)

    def test_fg(self):
        self.eq(self.orange('text'), '\033[38;5;214mtext\033[m')
        self.eq(self.coffee('text'), '\033[38;2;192;255;238mtext\033[m')
        self.eq(self.purple('text'), '\033[38;2;128;0;128mtext\033[m')
        self.eq(self.orange.fg('text'), '\033[38;5;214mtext\033[m')
        self.eq(self.coffee.fg('text'), '\033[38;2;192;255;238mtext\033[m')
        self.eq(self.purple.fg('text'), '\033[38;2;128;0;128mtext\033[m')

    def test_bg(self):
        self.eq(self.orange.bg('text'), '\033[48;5;214mtext\033[m')
        self.eq(self.coffee.bg('text'), '\033[48;2;192;255;238mtext\033[m')
        self.eq(self.purple.bg('text'), '\033[48;2;128;0;128mtext\033[m')

    def test_str(self):
        self.eq(str(self.orange), '\033[38;5;214m')
        self.eq(str(self.coffee), '\033[38;2;192;255;238m')
        self.eq(str(self.purple), '\033[38;2;128;0;128m')
        self.eq('{}'.format(self.orange), str(self.orange))
        self.eq('{}'.format(self.coffee), str(self.coffee))
        self.eq('{}'.format(self.purple), str(self.purple))

    def test_invert(self):
        self.eq(str(~self.orange), '\033[48;5;214m')
        self.eq(str(~self.coffee), '\033[48;2;192;255;238m')
        self.eq(str(~self.purple), '\033[48;2;128;0;128m')
        self.true(isinstance(~self.orange, ColorCompound))
        self.true(isinstance(~self.coffee, ColorCompound))
        self.true(isinstance(~self.purple, ColorCompound))

    def test_div(self):
        colors_under_test = (self.orange, self.coffee, self.purple)
        for A in colors_under_test:
            for B in colors_under_test:
                self.eq(A / B, paint(fg=A, bg=B))

        for A in colors_under_test:
            with self.raises(TypeError):
                A / 1

    def test_or(self):
        colors_under_test = (self.orange, self.coffee, self.purple)
        for A in colors_under_test:
            self.eq(nocolor | A, A)
            self.eq(A | nocolor, A)

        for A in colors_under_test:
            for B in colors_under_test:
                self.eq(A | B, B)


class TestColor256(TestCase):
    def test_color256_code(self):
        for i in range(256):
            self.eq(color(i).code, i)

    def test_color256_to_rgb(self):
        self.eq(color(0).to_rgb(), ColorRGB(0x00, 0x00, 0x00))
        self.eq(color(1).to_rgb(), ColorRGB(0x80, 0x00, 0x00))
        self.eq(color(2).to_rgb(), ColorRGB(0x00, 0x80, 0x00))
        self.eq(color(3).to_rgb(), ColorRGB(0x80, 0x80, 0x00))
        self.eq(color(4).to_rgb(), ColorRGB(0x00, 0x00, 0x80))
        self.eq(color(5).to_rgb(), ColorRGB(0x80, 0x00, 0x80))
        self.eq(color(6).to_rgb(), ColorRGB(0x00, 0x80, 0x80))
        self.eq(color(7).to_rgb(), ColorRGB(0xC0, 0xC0, 0xC0))
        self.eq(color(8).to_rgb(), ColorRGB(0x80, 0x80, 0x80))
        self.eq(color(9).to_rgb(), ColorRGB(0xFF, 0x00, 0x00))
        self.eq(color(10).to_rgb(), ColorRGB(0x00, 0xFF, 0x00))
        self.eq(color(11).to_rgb(), ColorRGB(0xFF, 0xFF, 0x00))
        self.eq(color(12).to_rgb(), ColorRGB(0x00, 0x00, 0xFF))
        self.eq(color(13).to_rgb(), ColorRGB(0xFF, 0x00, 0xFF))
        self.eq(color(14).to_rgb(), ColorRGB(0x00, 0xFF, 0xFF))
        self.eq(color(15).to_rgb(), ColorRGB(0xFF, 0xFF, 0xFF))
        self.eq(color(208).to_rgb(), ColorRGB(0xFF, 0x87, 0x00))
        self.eq(color(232).to_rgb(), ColorRGB(0x08, 0x08, 0x08))
        self.eq(color(237).to_rgb(), ColorRGB(0x3A, 0x3A, 0x3A))

    def test_color256_to_hsv(self):
        self.eq(color(0).to_hsv(), color(0).to_rgb().to_hsv())
        self.eq(color(1).to_hsv(), color(1).to_rgb().to_hsv())
        self.eq(color(2).to_hsv(), color(2).to_rgb().to_hsv())
        self.eq(color(3).to_hsv(), color(3).to_rgb().to_hsv())
        self.eq(color(4).to_hsv(), color(4).to_rgb().to_hsv())
        self.eq(color(5).to_hsv(), color(5).to_rgb().to_hsv())
        self.eq(color(6).to_hsv(), color(6).to_rgb().to_hsv())
        self.eq(color(7).to_hsv(), color(7).to_rgb().to_hsv())
        self.eq(color(8).to_hsv(), color(8).to_rgb().to_hsv())
        self.eq(color(9).to_hsv(), color(9).to_rgb().to_hsv())
        self.eq(color(10).to_hsv(), color(10).to_rgb().to_hsv())
        self.eq(color(11).to_hsv(), color(11).to_rgb().to_hsv())
        self.eq(color(12).to_hsv(), color(12).to_rgb().to_hsv())
        self.eq(color(13).to_hsv(), color(13).to_rgb().to_hsv())
        self.eq(color(14).to_hsv(), color(14).to_rgb().to_hsv())
        self.eq(color(15).to_hsv(), color(15).to_rgb().to_hsv())
        self.eq(color(208).to_hsv(), color(208).to_rgb().to_hsv())
        self.eq(color(232).to_hsv(), color(232).to_rgb().to_hsv())
        self.eq(color(237).to_hsv(), color(237).to_rgb().to_hsv())


class TestColorRGB(TestCase):
    def test_rgb_empty(self):
        self.eq(ColorRGB().seq, '')

    def test_rgb(self):
        some_color = ColorRGB(160.2, 90, 0)

        self.eq(some_color.r, 160.2)
        self.eq(some_color.g, 90)
        self.eq(some_color.b, 0)
        self.eq(some_color.rgb, (160.2, 90, 0))

        self.eq(some_color.R, 160)
        self.eq(some_color.G, 90)
        self.eq(some_color.B, 0)
        self.eq(some_color.RGB, (160, 90, 0))

        self.eq(int(some_color), 0xA05A00)

    def test_value_range_check(self):
        with self.raises(TypeError):
            ColorRGB(300, 300, 300)

    def test_rgb_mul(self):
        some_color = ColorRGB(160, 90, 0) * 2
        self.eq(some_color.r, 320)
        self.eq(some_color.g, 180)
        self.eq(some_color.b, 0)

        self.eq(some_color.R, 255)
        self.eq(some_color.G, 180)
        self.eq(some_color.B, 0)

        some_color = ColorRGB(160, 90, 0) * 0.8
        self.eq(str(some_color), '\033[38;2;128;72;0m')

    def test_rgb_div(self):
        some_color = ColorRGB(160, 90, 0) // 2
        self.eq(some_color, ColorRGB(80, 45, 0))

    def test_rgb_overflow(self):
        some_color = ColorRGB(160, 90, 0) * 2
        self.eq(str(some_color), '\033[38;2;255;180;0m')
        self.eq(int(some_color), 0xFFB400)

    def test_rgb_average(self):
        d_red = ColorRGB(255, 0, 0) * 2
        self.eq(d_red.r, 255 * 2)
        self.eq(d_red.g, 0)
        self.eq(d_red.b, 0)

        d_green = ColorRGB(0, 255, 0) * 2
        self.eq(d_green.r, 0)
        self.eq(d_green.g, 255 * 2)
        self.eq(d_green.b, 0)

        d_yellow = d_red + d_green
        self.eq(d_yellow.r, 255 * 2)
        self.eq(d_yellow.g, 255 * 2)
        self.eq(d_yellow.b, 0)

        yellow = d_yellow // 2
        self.eq(yellow, ColorRGB(255, 255, 0))

    def test_rgb_to_rgb(self):
        red = ColorRGB(255, 0, 0)
        self.true(red.to_rgb() is red)

    def test_rgb_to_hsv(self):
        red = ColorRGB(255, 0, 0)
        self.eq(red.to_hsv(), ColorHSV(0, 100, 100))

    def test_rgb_format(self):
        red = ColorRGB(255, 0, 0)
        self.eq('{:#x}'.format(red), '#ff0000')
        self.eq('{:#X}'.format(red), '#FF0000')
        with self.raises(TypeError):
            self.eq('{:d}'.format(red), str(red))


class TestColorHSV(TestCase):
    def test_copy_ctor(self):
        red = ColorHSV('@0,100,100')
        other = ColorHSV(red)
        self.eq(red, other)

    def test_hsv_empty(self):
        self.eq(ColorHSV().seq, '')

    def test_hsv_parse_str(self):
        self.eq(ColorHSV('@0,100,100'), ColorHSV(0, 100, 100))
        self.eq(ColorHSV('@41,100,100'), ColorHSV(41, 100, 100))
        self.eq(ColorHSV('@401,100,100'), ColorHSV(41, 100, 100))
        self.eq(repr(ColorHSV('@0,100,100')), 'ColorHSV(0deg, 100%, 100%)')
        self.eq(repr(ColorHSV('@360,100,100')), 'ColorHSV(0deg, 100%, 100%)')

    def test_to_rgb(self):
        self.eq(ColorHSV('@0,100,100').to_rgb(), ColorRGB(255, 0, 0))

    def test_to_hsv(self):
        red = ColorHSV('@0,100,100')
        self.true(red.to_hsv() is red)

    def test_hsv(self):
        some_color = ColorHSV(60, 90, 0)
        self.eq(some_color.h, 60)
        self.eq(some_color.s, 90)
        self.eq(some_color.v, 0)
        self.eq(int(some_color), 60090000)

    def test_value_range_check(self):
        with self.raises(TypeError):
            ColorHSV(0, 101, 0)

        with self.raises(TypeError):
            ColorHSV(0, 0, 101)

        with self.raises(TypeError):
            ColorHSV(0, 0, 101, 0)

    def test_hsv_mul(self):
        some_color = ColorHSV(180, 100, 100) * 2
        self.eq(some_color.h, 0)
        self.eq(some_color.s, 200)
        self.eq(some_color.v, 200)
        self.eq(str(some_color), '\033[38;2;255;0;0m')

        some_color = ColorHSV(0, 100, 100) * 0.8
        self.eq(str(some_color), '\033[38;2;204;41;41m')

    def test_hsv_div(self):
        some_color = ColorHSV(160, 90, 0) // 2
        self.eq(some_color, ColorHSV(80, 45, 0))

    def test_hsv_overflow(self):
        some_color = ColorHSV(160, 90, 100) * 2
        self.eq(str(some_color), '\033[38;2;255;0;170m')
        self.eq(int(some_color), 320100100)

    def test_hsv_average(self):
        red = ColorHSV(360, 100, 100)
        yellow = ColorHSV(60, 100, 100)

        d_orange = red + yellow
        self.eq(d_orange.h, 60)
        self.eq(d_orange.s, 200)
        self.eq(d_orange.v, 200)

        orange = d_orange // 2
        self.eq(orange, ColorHSV(30, 100, 100))

    def test_hsv_format(self):
        lime = ColorHSV(120, 100, 100)
        self.eq('{}'.format(lime), str(ColorRGB(0, 255, 0)))
        self.eq('{:#}'.format(lime), '(@120, 100%, 100%)')

        with self.raises(TypeError):
            '{:d}'.format(lime)


class TestBuiltInColors(TestCase):
    def test_nocolor(self):
        self.eq(nocolor(), '')
        self.eq(nocolor('text'), 'text')
        self.eq(str(nocolor), '\033[m')
        self.eq('{}'.format(nocolor), '\033[m')

    def test_str(self):
        self.eq(str(black),     '\033[38;5;0m')
        self.eq(str(maroon),    '\033[38;5;1m')
        self.eq(str(green),     '\033[38;5;2m')
        self.eq(str(olive),     '\033[38;5;3m')
        self.eq(str(navy),      '\033[38;5;4m')
        self.eq(str(purple),    '\033[38;5;5m')
        self.eq(str(teal),      '\033[38;5;6m')
        self.eq(str(silver),    '\033[38;5;7m')
        self.eq(str(grey),      '\033[38;5;8m')
        self.eq(str(red),       '\033[38;5;9m')
        self.eq(str(lime),      '\033[38;5;10m')
        self.eq(str(yellow),    '\033[38;5;11m')
        self.eq(str(blue),      '\033[38;5;12m')
        self.eq(str(fuchsia),   '\033[38;5;13m')
        self.eq(str(magenta),   '\033[38;5;13m')
        self.eq(str(cyan),      '\033[38;5;14m')
        self.eq(str(aqua),      '\033[38;5;14m')
        self.eq(str(white),     '\033[38;5;15m')
        self.eq(str(darkorange),'\033[38;5;208m')
        self.eq(str(murasaki),  '\033[38;5;135m')

    def test_invert(self):
        self.eq(~red,      paint(bg=red))
        self.eq(~green,    paint(bg=green))
        self.eq(~yellow,   paint(bg=yellow))
        self.eq(~blue,     paint(bg=blue))
        self.eq(~magenta,  paint(bg=magenta))
        self.eq(~cyan,     paint(bg=cyan))
        self.eq(~white,    paint(bg=white))
        self.eq(~orange,   paint(bg=orange))

    def test_call(self):
        self.eq(black('text'),     '\033[38;5;0mtext\033[m')
        self.eq(maroon('text'),    '\033[38;5;1mtext\033[m')
        self.eq(green('text'),     '\033[38;5;2mtext\033[m')
        self.eq(olive('text'),     '\033[38;5;3mtext\033[m')
        self.eq(navy('text'),      '\033[38;5;4mtext\033[m')
        self.eq(purple('text'),    '\033[38;5;5mtext\033[m')
        self.eq(teal('text'),      '\033[38;5;6mtext\033[m')
        self.eq(silver('text'),    '\033[38;5;7mtext\033[m')
        self.eq(grey('text'),      '\033[38;5;8mtext\033[m')
        self.eq(red('text'),       '\033[38;5;9mtext\033[m')
        self.eq(lime('text'),      '\033[38;5;10mtext\033[m')
        self.eq(yellow('text'),    '\033[38;5;11mtext\033[m')
        self.eq(blue('text'),      '\033[38;5;12mtext\033[m')
        self.eq(fuchsia('text'),   '\033[38;5;13mtext\033[m')
        self.eq(magenta('text'),   '\033[38;5;13mtext\033[m')
        self.eq(cyan('text'),      '\033[38;5;14mtext\033[m')
        self.eq(aqua('text'),      '\033[38;5;14mtext\033[m')
        self.eq(white('text'),     '\033[38;5;15mtext\033[m')
        self.eq(darkorange('text'),'\033[38;5;208mtext\033[m')
        self.eq(murasaki('text'),  '\033[38;5;135mtext\033[m')


class TestPaint(TestCase):
    def test_repr(self):
        self.true(repr(paint()).startswith('ColorCompound'))

    def test_or(self):
        self.eq(black | (~yellow), paint(fg=0, bg=11))

        ry = red / yellow
        ig = ~green
        ryig = ry | ig
        self.eq(ryig, paint(fg=red, bg=green))

    def test_div(self):
        ry = red / yellow
        bg = blue / green
        rybg = ry / bg
        self.eq(rybg, paint(fg=red, bg=blue))
        self.eq(rybg('text'), '\033[38;5;9;48;5;12mtext\033[m')

    def test_invert(self):
        ry = red / yellow
        bg = blue / green
        rybg = ry / bg
        self.eq(~rybg, paint(fg=blue, bg=red))
        self.eq((~rybg)('text'), '\033[38;5;12;48;5;9mtext\033[m')


class TestDecolor(TestCase):
    def test_decolor(self):
        self.eq(decolor(orange('test')), 'test')
        self.eq(decolor('\033[1;31mred\033[m'), 'red')


class TestGradient(TestCase):
    def test_invalid_values(self):
        with self.raises(TypeError):
            gradient(True, False)

        A = color()
        B = color()

        with self.raises(TypeError):
            gradient(A, B, 1.5)

        with self.raises(ValueError):
            gradient(A, B, 1)

    def test_trivial(self):
        # N=2 trivial case
        A = color(39)
        B = color(214)
        self.eq(gradient(A, B, 2), (A, B))

        # Color256() and ColorRGB() case
        A = color(39)
        B = color('#C0FFEE')
        self.eq(gradient(A, B), (A, B))

        # Color256() rgb6 and gray case
        A = color(39)
        B = color(255)
        self.eq(gradient(A, B), (A, B))

    def test_color256_gray(self):
        A = color(235)
        B = color(245)

        # default length
        res = gradient(A, B)
        ans = tuple(range(235, 246))
        self.eq(res, tuple(map(color, ans)))

        # shorter length
        res = gradient(A, B, N=5)
        ans = (235, 238, 241, 243, 245)
        self.eq(res, tuple(map(color, ans)))

        # longer length
        res = gradient(A, B, N=15)
        ans = (235, 235, 236, 236, 237, 237, 238, 238, 239, 240, 241, 242, 243, 244, 245)
        self.eq(res, tuple(map(color, ans)))


    def test_color256_rgb(self):
        A = color(39)
        B = color(214)

        # default length
        res = gradient(A, B)
        ans = (39 ,74 ,109 ,144 ,179 ,214)
        self.eq(res, tuple(map(color, ans)))

        # shorter length
        res = gradient(A, B, N=4)
        ans = (39, 109, 179, 214)
        self.eq(res, tuple(map(color, ans)))

        # longer length
        res = gradient(A, B, N=15)
        ans = (39, 39, 39, 74, 74, 74, 109, 109, 109, 144, 144, 179, 179, 214, 214)
        self.eq(res, tuple(map(color, ans)))

    def test_rgb(self):
        A = color(242, 5, 148)
        B = color(146, 219, 189)

        # default length
        res = gradient(A, B)
        ans = (ColorRGB(242, 5, 148),
               ColorRGB(218, 58, 158),
               ColorRGB(194, 112, 168),
               ColorRGB(170, 166, 179),
               ColorRGB(146, 219, 189))
        for i in range(len(res)):
            self.eq(res[i], ans[i])

        # shorter length
        res = gradient(A, B, N=4)
        ans = (ColorRGB(242, 5, 148),
               ColorRGB(210, 76, 162),
               ColorRGB(178, 148, 175),
               ColorRGB(146, 219, 189))
        for i in range(len(res)):
            self.eq(res[i], ans[i])

        # longer length
        res = gradient(A, B, N=15)
        ans = (ColorRGB(242, 5, 148),
               ColorRGB(235, 20, 151),
               ColorRGB(228, 36, 154),
               ColorRGB(221, 51, 157),
               ColorRGB(215, 66, 160),
               ColorRGB(208, 81, 163),
               ColorRGB(201, 97, 166),
               ColorRGB(194, 112, 168),
               ColorRGB(187, 127, 171),
               ColorRGB(180, 143, 174),
               ColorRGB(173, 158, 177),
               ColorRGB(167, 173, 180),
               ColorRGB(160, 188, 183),
               ColorRGB(153, 204, 186),
               ColorRGB(146, 219, 189))
        for i in range(len(res)):
            self.eq(res[i], ans[i])

        A = color('#FF1100')
        B = color('#FF0011')
        res = gradient(A, B, N=3)
        self.eq(res, (A, color('#FF0808'), B))

    def test_hsv(self):
        A = ColorHSV(0, 100, 100)
        B = ColorHSV(300, 50, 100)

        # clockwise
        res = gradient(A, B, clockwise=True)
        res2 = gradient(A, B)
        ans = (ColorHSV(0, 100, 100),
               ColorHSV(33, 94, 100),
               ColorHSV(66, 88, 100),
               ColorHSV(100, 83, 100),
               ColorHSV(133, 77, 100),
               ColorHSV(166, 72, 100),
               ColorHSV(200, 66, 100),
               ColorHSV(233, 61, 100),
               ColorHSV(266, 55, 100),
               ColorHSV(300, 50, 100),)

        self.eq(len(res), len(ans))
        for a, b, c in zip(res, res2, ans):
            # Check if the colors are close enough
            self.le(abs(sum(a.hsv) - sum(b.hsv)), 2)
            self.le(abs(sum(a.hsv) - sum(c.hsv)), 2)

        res = gradient(A, B, clockwise=True, N=5)
        res2 = gradient(A, B)
        ans = (ColorHSV(0, 100, 100),
               ColorHSV(75, 88, 100),
               ColorHSV(150, 75, 100),
               ColorHSV(225, 62, 100),
               ColorHSV(300, 50, 100),)

        for a, b in zip(res, ans):
            # Check if the colors are close enough
            self.le(abs(sum(a.hsv) - sum(b.hsv)), 2)

        # counter-clockwise
        res = gradient(A, B, clockwise=False)
        ans = (ColorHSV(0, 100, 100),
               ColorHSV(345, 87, 100),
               ColorHSV(330, 75, 100),
               ColorHSV(315, 62, 100),
               ColorHSV(300, 50, 100),)

        self.eq(len(res), len(ans))
        for a, b in zip(res, ans):
            # Check if the colors are close enough
            self.le(abs(sum(a.hsv) - sum(b.hsv)), 2)

        # reverse
        res = gradient(A, B, clockwise=True)
        rev = gradient(A, B, clockwise=True, reverse=True)

        for a, b in zip(res, rev[::-1]):
            # Check if the colors are close enough
            self.le(abs(sum(a.hsv) - sum(b.hsv)), 2)

        # reverse
        res = gradient(A, B, clockwise=False)
        rev = gradient(A, B, clockwise=False, reverse=True)

        for a, b in zip(res, rev[::-1]):
            # Check if the colors are close enough
            self.le(abs(sum(a.hsv) - sum(b.hsv)), 2)



import sys
import shutil
import argparse
import textwrap

from os.path import basename

from . import lib_colors

from .lib_colors import paint
from .lib_colors import color
from .lib_regex import rere
from .lib_math import resample
from .lib_math import is_uint8
from .lib_itertools import lookahead
from .lib_tui import getch


errors = []
def pend_error(*errmsg):
    errors.append(errmsg)


def judge_errors():
    if errors:
        for error in errors:
            print(*error)
        sys.exit(1)


def high_contrast_fg(c):
    if isinstance(c, lib_colors.Color256):
        c = c.index

    if c == 0:
        return None

    if c < 16:
        return 0

    elif c < 232:
        base = c - 16
        r = (base // 36)
        g = ((base % 36) // 6)
        b = (base % 6)

        if r < 2 and g < 2:
            return None

        if r < 3 and g < 1:
            return None

    elif c in range(232, 240):
        return None

    return 0


def parse_target(arg):
    if not isinstance(arg, str):
        return

    ret = None

    to = []

    while True:
        m = rere(arg)

        if m.fullmatch(r'^(.+)\.(rgb|RGB|hsv|HSV)$'):
            to.append(m.group(2))
            arg = m.group(1)
            continue

        break

    if arg in lib_colors.names:
        ret = getattr(lib_colors, arg)

    # #RRGGBB format
    elif m.fullmatch(r'#?([0-9a-fA-Z]{6})'):
        ret = color('#' + m.group(1))

    # #RRR,GGG,BBB format
    elif m.fullmatch(r'#([0-9]+),([0-9]+),([0-9]+)'):
        r, g, b = map(lambda x: int(x, 10), arg[1:].split(','))
        ret = lib_colors.ColorRGB(r, g, b)

    # @HHH,SSS,VVV format
    elif m.fullmatch(r'@([0-9]+),([0-9]+),([0-9]+)'):
        ret = lib_colors.ColorHSV(arg)

    # int
    elif m.fullmatch(r'[0-9]+'):
        try:
            i = int(arg, 10)
            if is_uint8(i):
                ret = color(i)
        except:
            ret = None

    tr_path = arg
    for t in to[::-1]:
        try:
            if t.lower() == 'rgb':
                ret = ret.to_rgb()
                tr_path += '.rgb'
            elif t.lower() == 'hsv':
                ret = ret.to_hsv()
                tr_path += '.hsv'
        except AttributeError:
            pend_error('Error: Cannot transform color', tr_path, 'to', t)

    return ret


def spell_suggestions(word):
    import difflib
    return difflib.get_close_matches(word, lib_colors.names, cutoff=0)


def spell_suggestion_err_msg(word):
    err_msg = 'Unknown color name "{}"'.format(word)
    suggestions = spell_suggestions(word)[:3]
    if suggestions:
        err_msg += ', did you mean '
        if len(suggestions) == 1:
            err_msg += '"{}"'.format(suggestions[0])
        elif len(suggestions) == 2:
            err_msg += '"{}", or "{}"'.format(suggestions[0], suggestions[1])
        elif len(suggestions) == 3:
            err_msg += '"{}", "{}", or "{}"'.format(*suggestions)
        err_msg += '?'
    pend_error(err_msg)


def expand_macro_all():
    ret = []
    for i in range(256):
        ret.append((parse_target(str(i)), []))

    for name in lib_colors.names:
        c = parse_target(name)
        ret[c.index][1].append(name)

    return ret


def expand_macro_named():
    ret = []
    for name in lib_colors.names:
        ret.append((parse_target(name), [name]))
    return ret


def main_256cube():
    # Print color cube palette
    print('Format: ESC[30;48;5;{}m')
    for c in range(0, 256):
        color_rgb = color(c).to_rgb()
        print(paint(fg=high_contrast_fg(c), bg=c)(' ' + str(c).rjust(3)), end='')

        if c < 16 and (c + 1) % 8 == 0:
            print()
        if c >= 16 and (c - 16 + 1) % 36 == 0:
            print()
        if c in (15, 231):
            print()

    print()
    sys.exit()


def main():
    prog = basename(sys.argv[0])
    argv = sys.argv[1:]

    if not argv:
        main_256cube()

    colorful = ''.join(
            map(
                lambda x: lib_colors.color(x[0])(x[1]),
                zip(
                    ['#FF0000', '#FFC000', '#FFFF00',
                     '#C0FF00', '#00FF00', '#00FFC0',
                     '#00FFFF', '#00C0FF', '#3333FF', '#C000FF', '#FF00FF'],
                    'coooolorful'
                    )
                )
            )
    parser = argparse.ArgumentParser(
            prog=prog,
            description=('Query pre-defined colors from warawara, ' +
                         'or produce ' + colorful + ' strips/tiles to fill the screen.'),
            epilog=textwrap.dedent('''
                    Example usages:
                    $ {prog}
                    $ {prog} all
                    $ {prog} named --grep orange --hex
                    $ {prog} FFD700 --rgb
                    $ {prog} --tile --lines=2 --cols=8 salmon white
                    $ {prog} --gradient 00AFFF FFAF00
                    '''.strip('\n')).format(prog=prog),
            allow_abbrev=False, add_help=False,
            formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-h', '--help', action='help', help='Show this help message and exit')

    parser.add_argument('--grep',
                        action='append',
                        help='''Filter out colors that does not contain the specified sub-string
This argument can be specified multiple times for multiple keywords''')

    parser.add_argument('-a', '--aliases',
                        action='store_true',
                        help='Show aliases of specified colors')

    parser.add_argument('--hex', dest='val_fmt',
                        action='append_const', const='hex',
                        help='Show RGB value in hex number')

    parser.add_argument('--rgb', dest='val_fmt',
                        action='append_const', const='rgb',
                        help='Show RGB value in 3-tuple')

    parser.add_argument('--hsv', dest='val_fmt',
                        action='append_const', const='hsv',
                        help='Show HSV value in 3-tuple')

    parser.add_argument('--sort',
                        nargs='?', choices=['index', 'name', 'rgb', 'hue', 'no'], const='index',
                        default='no',
                        help='Sort the output by the specified attribute')

    parser.add_argument('-r', '--reverse',
                        action='store_true',
                        help='''Reverse the output sequence''')

    class YesNoToBoolOption(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, self.dest, values == 'yes')
    parser.add_argument('-m', '--merge',
                        action=YesNoToBoolOption, nargs='?', choices=['yes', 'no'], const='yes',
                        help='Merge colors that have same index')

    parser.add_argument('-M', '--no-merge',
                        action='store_false', dest='merge',
                        help='Dont merge colors that have same index')

    parser.add_argument('-t', '--tile',
                        action='store_true',
                        help='''Tiles to fill the whole screen
Ignores every other optional arguments except for --height and --width
Ignores "all" and "named" macros''')

    parser.add_argument('-g', '--gradient',
                        action='store_true',
                        help='''Calculate color gradient for specified two colors
Ignores "all" and "named" macros''')

    parser.add_argument('-c', '--clockwise',
                        action=YesNoToBoolOption, nargs='?', choices=['yes', 'no'], const='yes',
                        help='''Calculate clockwise color gradient for HSV''')

    parser.add_argument('--cols', '--columns',
                        type=int,
                        help='Specify terminal columns')

    parser.add_argument('--lines',
                        type=int,
                        help='Specify terminal height')

    parser.add_argument('targets', nargs='*', help='''Names / indexs / RGB hex values / HSV values to query
"all" and "named" macros could be used in "list" mode''')

    parser.set_defaults(val_fmt=[])

    args = parser.parse_intermixed_args()

    if args.tile and args.gradient:
        print('--tile and --gradient cannot be used together')
        sys.exit(1)

    elif args.tile:
        main_tile(args)

    else:
        main_list(args)


def main_list(args):
    if args.gradient:
        # argument handling for gradient
        if not args.targets:
            pend_error('No colors to gradient')

        if len(args.targets) == 1:
            pend_error('Need destination color')

        if len(args.targets) > 3:
            pend_error('Too many arguments')

        judge_errors()

        arg_src = args.targets[0]
        arg_dst = args.targets[1]

        if len(args.targets) == 3:
            arg_n = args.targets[2]
        else:
            arg_n = None

        src = parse_target(arg_src)
        if not src:
            spell_suggestion_err_msg(arg_src)

        dst = parse_target(arg_dst)
        if not dst:
            spell_suggestion_err_msg(arg_dst)

        try:
            n = int(arg_n, 10) if arg_n else None
        except:
            pend_error('Invalid number: {}'.format(arg_n))

        judge_errors()

        def color_text(this_color):
            if isinstance(this_color, lib_colors.Color256):
                return str(this_color.index)
            elif isinstance(this_color, lib_colors.ColorRGB):
                return '{:#X}'.format(this_color)
            elif isinstance(this_color, lib_colors.ColorHSV):
                return '@{:},{:},{:}'.format(this_color.H, this_color.S, this_color.V)
            else:
                line.append('(?)')

        expanded = [(g, color_text(g)) for g in lib_colors.gradient(src, dst, n, reverse=args.reverse, clockwise=args.clockwise)]

    else:
        # argument handling for not gradient
        if args.merge is None:
            if 'all' in args.targets or 'named' in args.targets:
                args.merge = True
            elif args.aliases:
                args.merge = True
            else:
                args.merge = False

        if args.grep and not args.targets:
            args.targets = ['all']

        expanded = []
        for arg in args.targets:
            if arg in ('all', 'named'):
                if arg == 'all':
                    expansion = expand_macro_all()
                elif arg == 'named':
                    expansion = expand_macro_named()

                for entry in expansion:
                    if not entry[1]:
                        expanded.append((entry[0], None))
                    else:
                        for name in entry[1]:
                            expanded.append((entry[0], name))
                continue

            t = parse_target(arg)
            if t:
                expanded.append((t, arg))
            else:
                spell_suggestion_err_msg(arg)

        judge_errors()

    # Grep
    if args.grep:
        for keyword in args.grep:
            tmp, expanded = expanded, []
            for C, name in tmp:
                if keyword == '':
                    if not name:
                        expanded.append((C, name))
                else:
                    if name and keyword in name:
                        expanded.append((C, name))
            del tmp

    # Sort
    if args.sort == 'index':
        expanded.sort(key=lambda x: int(x[0]) + (isinstance(x[0], lib_colors.ColorRGB) << 8))
    elif args.sort == 'name':
        expanded.sort(key=lambda x: ((isinstance(x[0], lib_colors.ColorRGB)), x[1]))
    elif args.sort == 'rgb':
        expanded.sort(key=lambda x: (x[0].to_rgb() if isinstance(x[0], lib_colors.Color256) else x[0]).rgb)
    elif args.sort == 'hue':
        def to_hsv(x):
            if isinstance(x, lib_colors.ColorHSV):
                return x
            if isinstance(x, lib_colors.ColorRGB):
                return x.to_hsv()
            if isinstance(x, lib_colors.Color256):
                return x.to_rgb().to_hsv()
        expanded.sort(key=lambda x: to_hsv(x[0]).h)

    inventory = []
    def stage(color, name):
        if name is None:
            name = ''

        m = rere(name)
        if not args.gradient and m.fullmatch(r'[0-9]+') and not m.fullmatch(r'#?[0-9a-fA-F]{6}'):
            name = ''

        if not args.merge:
            inventory.append((color, [name] if name else []))
            return

        entries = filter(lambda entry: entry[0] == color, inventory)

        try:
            entry = next(entries)
            append = False
            if not entry[1]:
                append = True
            if name and args.merge and name not in entry[1]:
                append = True

            if append:
                entry[1].append(name)

            return

        except StopIteration:
            pass

        inventory.append((color, [name] if name else []))

    for entry in expanded:
        stage(entry[0], entry[1])

    if not inventory:
        print('No colors to query')
        sys.exit(1)

    unmentioned_names = set(lib_colors.names)
    for _, names in inventory:
        unmentioned_names -= set(names)

    aliases = [[] for i in range(256)]
    if args.aliases:
        for name in unmentioned_names:
            c = getattr(lib_colors, name).index
            aliases[c].append(name)

    for this_color, names in inventory:
        line = []
        rgb = this_color if isinstance(this_color, lib_colors.ColorRGB) else this_color.to_rgb()
        hsv = rgb.to_hsv()

        if isinstance(this_color, lib_colors.Color256):
            line.append('{:>3}'.format(this_color.index))
        elif isinstance(this_color, lib_colors.ColorRGB):
            line.append('(#)')
        elif isinstance(this_color, lib_colors.ColorHSV):
            line.append('(@)')
        else:
            line.append('(?)')

        for val_fmt in args.val_fmt:
            if val_fmt == 'rgb':
                line.append('({:>3}, {:>3}, {:>3})'.format(*rgb.RGB))

            elif val_fmt == 'hex':
                line.append('{:#X}'.format(rgb))

            elif val_fmt == 'hsv':
                line.append('(@{:>3}, {:>3}%, {:>3}%)'.format(*hsv.HSV))

        line.append(paint(fg=this_color, bg=this_color)('warawara'))

        line.append(', '.join(names))

        if isinstance(this_color, lib_colors.Color256):
            aliases[this_color.index] = [name
                                         for name in aliases[this_color.index]
                                         if name not in names]
            if aliases[this_color.index]:
                a = ('(' + ', '.join(aliases[this_color.index]) + ')')
                line[-1] = line[-1] + (' ' if line[-1] else '') + a
                aliases[this_color.index] = []

        print(' '.join(line))


def main_tile(args):
    if not args.targets:
        pend_error('No colors to tile')

    judge_errors()

    tiles = [[]]
    for arg in args.targets:
        for token in arg.split('/'):
            if token in ('all', 'named'):
                pend_error('"{}" cannot be used in tile mode'.format(token))
                continue

            t = parse_target(token)
            if not t:
                spell_suggestion_err_msg(token)
                continue

            tiles[-1].append((token, t))

        tiles.append([])

    judge_errors()

    if not tiles[-1]:
        tiles.pop()

    cols, lines = shutil.get_terminal_size()
    cols = args.cols or cols
    lines = args.lines or lines

    if lines < 0:
        lines = len(tiles)

    for idx, is_last in lookahead(resample(range(len(tiles)), lines)):
        colors = tiles[idx]
        widths = []
        quo, rem = divmod(cols, len(colors))
        widths = [quo + (i < rem) for i, elem in enumerate(colors)]
        line = ''
        for idx, textcolor in enumerate(colors):
            text, c = textcolor
            text = text[:widths[idx]]
            line += paint(fg=c, bg=c)(text) + (~c)(' ' * (widths[idx] - len(text)))

        if args.lines:
            print(line)
        else:
            print(line, end='\n' if not is_last else '')

    sys.stdout.flush()

    getch()



import unittest.mock

from .lib_test_utils import *

import warawara as wara


class TestOpen(TestCase):
    def test_open_wb(self):
        mock_open = unittest.mock.mock_open()
        self.patch('builtins.open', mock_open)
        with wara.open('warawara.txt', 'wb') as f:
            mock_open.assert_called_once_with('warawara.txt', mode='wb')

    def test_open_rb(self):
        mock_open = unittest.mock.mock_open()
        self.patch('builtins.open', mock_open)
        with wara.open('warawara.txt', 'rb') as f:
            mock_open.assert_called_once_with('warawara.txt', mode='rb')

    def test_open_without_mode(self):
        mock_open = unittest.mock.mock_open()
        self.patch('builtins.open', mock_open)
        with wara.open('warawara.txt') as f:
            mock_open.assert_called_once_with(
                    'warawara.txt', mode='rt',
                    encoding='utf-8', errors='backslashreplace')

    def test_write(self):
        mock_open = unittest.mock.mock_open()
        self.patch('builtins.open', mock_open)
        with wara.open('warawara.txt', 'w') as f:
            mock_open.assert_called_once_with(
                    'warawara.txt', mode='w',
                    encoding='utf-8', errors='backslashreplace')

            f.write('normal write\n')
            f.writeline()
            f.writeline('line1')
            f.writelines(['line2', 'line3', 'line4'])

        handle = mock_open()
        handle.write.assert_has_calls([
            unittest.mock.call('normal write\n'),
            unittest.mock.call('\n'),
            unittest.mock.call('line1\n'),
            unittest.mock.call('line2\n'),
            unittest.mock.call('line3\n'),
            unittest.mock.call('line4\n'),
            ])

    def test_read(self):
        answer = [
                'line1',
                'line2',
                'line3',
                'and line4',
                'and line5',
                ]
        mock_open = unittest.mock.mock_open(read_data='\n'.join(answer))
        self.patch('builtins.open', mock_open)

        lines = []
        with wara.open('warawara.txt', 'r') as f:
            mock_open.assert_called_once_with(
                    'warawara.txt', mode='r',
                    encoding='utf-8', errors='backslashreplace')

            lines.append(f.readline())
            lines.append(f.readline())
            lines += f.readlines()

        self.eq(lines, answer)

    def test_read_without_rstrip(self):
        answer = [
                'line1',
                'line2',
                'line3',
                'and line4',
                'and line5',
                ]
        mock_open = unittest.mock.mock_open(read_data='\n'.join(answer))
        self.patch('builtins.open', mock_open)

        lines = []
        with wara.open('warawara.txt', mode='r', rstrip='') as f:
            mock_open.assert_called_once_with(
                    'warawara.txt', mode='r',
                    encoding='utf-8', errors='backslashreplace')

            lines.append(f.readline())
            lines.append(f.readline())
            lines += f.readlines()

        self.eq(lines, [
                'line1\n',
                'line2\n',
                'line3\n',
                'and line4\n',
                'and line5',
            ])


class TestNatsorted(TestCase):
    def test_natsorted(self):
        self.eq(
                wara.natsorted([
                    'apple1',
                    'apple10',
                    'banana10',
                    'apple2',
                    'banana1',
                    'banana3',
                    ]),
                [
                    'apple1',
                    'apple2',
                    'apple10',
                    'banana1',
                    'banana3',
                    'banana10',
                    ])

        self.eq(
                wara.natsorted([
                    'version-1.9',
                    'version-2.0',
                    'version-1.11',
                    'version-1.10',
                    ]),
                [
                    'version-1.9',
                    'version-1.10',
                    'version-1.11',
                    'version-2.0',
                    ])

        from pathlib import Path
        self.eq(
                wara.natsorted([
                    Path('version-1.9'),
                    Path('version-2.0'),
                    Path('version-1.11'),
                    Path('version-1.10'),
                    ]),
                [
                    Path('version-1.9'),
                    Path('version-1.10'),
                    Path('version-1.11'),
                    Path('version-2.0'),
                    ])



from .lib_test_utils import *

import warawara as wara


class TestMath(TestCase):
    def test_sgn(self):
        sgn = wara.sgn
        self.eq(sgn(2024), 1)
        self.eq(sgn(0), 0)
        self.eq(sgn(-2024), -1)

    def test_lerp(self):
        lerp = wara.lerp
        self.eq(lerp(0, 10, 0), 0)
        self.eq(lerp(0, 10, 0.5), 5)
        self.eq(lerp(0, 10, 1), 10)
        self.eq(lerp(0, 10, 2), 20)

        self.eq(lerp(0, 9, 0.333333333333333333), 3)
        self.eq(lerp(-10, 10, 0), -10)
        self.eq(lerp(-10, 10, 0.1), -8)
        self.eq(lerp(-10, 10, 0.5), 0)
        self.eq(lerp(-10, 10, 1), 10)

    def test_clamp(self):
        clamp = wara.clamp
        self.eq(clamp(3, 0, 7), 3)
        self.eq(clamp(3, 5, 7), 5)
        self.eq(clamp(3, 9, 7), 7)

        self.eq(clamp(7, 0, 3), 3)
        self.eq(clamp(7, 5, 3), 5)
        self.eq(clamp(7, 9, 3), 7)

    def test_vector(self):
        vector = wara.vector
        v1 = vector(1, 2, 3)
        v2 = vector([4, 5, 6])

        self.eq(v1, vector(v1))

        self.eq(v1 + 2, (3, 4, 5))
        self.eq(2 + v1, (3, 4, 5))
        self.eq(v1 + v2, (5, 7, 9))
        self.eq(v1 + (4, 5, 6), (5, 7, 9))

        self.eq(v1 - 2, (-1, 0, 1))
        self.eq(v1 - v2, (-3, -3, -3))
        self.eq(v1 - (4, 5, 6), (-3, -3, -3))

        self.eq(v1 * 2, (2, 4, 6))
        self.eq(2 * v1, (2, 4, 6))
        self.eq(v1 * (4, 5, 6), (4, 10, 18))

        self.eq(v1 / 2, (0.5, 1.0, 1.5))
        self.eq(v1 // 2, (0, 1, 1))

        with self.raises(TypeError):
            v1 / v2

        with self.raises(TypeError):
            v1 // v2

        self.eq(v1.map(lambda x: 3 * x + 1), (4, 7, 10))

        with self.raises(ValueError):
            vector('bla')

        with self.raises(TypeError):
            vector(1, 2, 3) == 'bla'

        self.eq(repr(v1), '(1, 2, 3)')

        with self.raises(ValueError):
            v1 + (1, 2, 3, 4)

        with self.raises(ValueError):
            v1 - (1, 2, 3, 4)

        with self.raises(ValueError):
            v1 * (1, 2, 3, 4)

        lerp = wara.lerp
        self.eq(lerp(v1, v2, 0), v1)
        self.eq(lerp(v1, v2, 0.5), (2.5, 3.5, 4.5))
        self.eq(lerp(v1, v2, 1), v2)
        self.eq(lerp(v1, v2, 2), (7, 8, 9))

    def test_interval(self):
        interval = wara.interval
        self.eq(interval(1, 3), [1, 2, 3])
        self.eq(interval(3, 1), [3, 2, 1])
        self.eq(interval(3, -3), [3, 2, 1, 0, -1, -2, -3])
        self.eq(interval(3, -3, close=False), [2, 1, 0, -1, -2])

        self.eq(interval(3, 3), [3])
        self.eq(interval(3, 3, close=False), [])

    def test_resample(self):
        resample = wara.resample

        samples = (1, 2, 3, 4, 5)

        self.eq(resample(samples, 5), samples)
        self.eq(resample(samples, 3), (1, 3, 5))
        self.eq(resample(samples, 10), (1, 1, 2, 2, 3, 3, 4, 4, 5, 5))



import sys

import unittest.mock

from .lib_test_utils import *

import warawara as wara


class TestBinWara(TestCase):
    def setUp(self):
        self.prints = []
        self.patch('builtins.print', self.mock_print)

    def tearDown(self):
        pass

    def mock_print(self, *args, **kwargs):
        self.prints.append((args, kwargs))

    @property
    def stdout(self):
        return [' '.join(args)
                for args, kwargs in self.prints
                if kwargs.get('file', sys.stdout) == sys.stdout]

    @property
    def stderr(self):
        return [' '.join(args)
                for args, kwargs in self.prints
                if kwargs.get('file') == sys.stderr]

    def test_bin_wara(self):
        # Run 'warawara' and check output
        with self.raises(SystemExit):
            sys.argv = ['warawara']
            wara.bin.wara.main()

        modules = self.stdout

        import os
        import re
        files = list(
                map(
                    lambda x: re.fullmatch(r'bin_(\w+).py', x).group(1),
                    filter(lambda x: x.startswith('bin') and x.endswith('.py'),
                           os.listdir('warawara'))
                    )
                )
        self.eq(sorted(modules), sorted(files))

    def test_bin_wara_subcmd_rainbow(self):
        with self.raises(SystemExit):
            sys.argv = ['warawara', 'rainbow']
            wara.bin.wara.main()

        sys.argv = ['warawara', 'rainbow', 'murasaki']
        wara.bin.wara.main()

    def test_bin_wara_subcmd_unknown(self):
        with self.raises(SystemExit):
            sys.argv = ['warawara', 'wow']
            wara.bin.wara.main()

        self.true('Unknown' in '\n'.join(self.stderr))

    def test_bin_wara_recursion_error(self):
        with self.raises(SystemExit):
            sys.argv = ['warawara', 'wara', 'wara', 'wara']
            wara.bin.wara.main()

        self.true('RecursionError' in '\n'.join(self.stderr))



import threading

from .lib_test_utils import *

import warawara as wara

class TestRunInThread(TestCase):
    def test_run_in_thread(self):
        barrier = threading.Barrier(2)
        checkpoint = False

        def may_stuck():
            nonlocal checkpoint
            barrier.wait()
            checkpoint = True

        with self.run_in_thread(may_stuck):
            self.false(checkpoint)
            barrier.wait()

        self.true(checkpoint)

    def test_run_in_thread_reuse(self):
        def foo():
            pass

        p = self.run_in_thread(foo)
        with p:
            pass

        with self.raises(RuntimeError):
            with p:
                pass


class TestCheckPoint(TestCase):
    def test_checkpoint(self):
        Checkpoint = wara.Checkpoint

        checkpoint = Checkpoint(self)

        self.false(checkpoint)
        checkpoint.set()
        checkpoint.check()
        checkpoint.check(True)
        self.true(checkpoint)

        checkpoint.clear()
        checkpoint.check(False)
        self.false(checkpoint)

        def set_checkpoint():
            checkpoint.wait()

        with self.run_in_thread(set_checkpoint):
            checkpoint.set()

        checkpoint.check()


class TestSubprocRunMocker(TestCase):
    def test_mock_basic(self):
        mock_run = RunMocker()
        def mock_wah(proc, *args):
            proc.stdout.writeline('mock wah')
            if args:
                proc.stdout.writeline(' '.join(args))
            return 0
        mock_run.register('wah', mock_wah)

        p = mock_run('wah'.split())
        self.eq(p.stdout.lines, ['mock wah'])

        p = mock_run('wah')
        self.eq(p.stdout.lines, ['mock wah'])

        p = mock_run('wah wah wah'.split())
        self.eq(p.stdout.lines, ['mock wah', 'wah wah'])

    def test_mock_meaningless_mock(self):
        mock_run = RunMocker()
        with self.raises(ValueError):
            mock_run.register('cmd')

    def test_mock_ambiguous_mock(self):
        mock_run = RunMocker()
        with self.raises(ValueError):
            mock_run.register('wah', lambda: None, stdout='wah')

    def test_mock_cmd_with_wrong_type(self):
        mock_run = RunMocker()
        with self.raises(ValueError):
            mock_run.register(['wah'], lambda proc: 0)

    def test_mock_callback_with_wrong_type(self):
        mock_run = RunMocker()
        with self.raises(TypeError):
            mock_run.register('wah', 0)

    def test_mock_run_empty_cmd(self):
        mock_run = RunMocker()
        with self.raises(ValueError):
            mock_run([])

    def test_mock_run_unregistered_cmd(self):
        mock_run = RunMocker()
        with self.raises(ValueError):
            p = mock_run('ls -a -l'.split())

    def test_mock_register_default_cmd(self):
        mock_run = RunMocker()
        def default_cmd(proc, *args):
            self.eq(args, ('-a', '-l'))
            return 42
        mock_run.register('*', default_cmd)
        p = mock_run('ls -a -l'.split())
        self.eq(p.returncode, 42)

    def test_mock_with_returncode(self):
        mock_run = RunMocker()
        mock_run.register('wah', returncode=1)
        p = mock_run(['wah'])
        self.eq(p.returncode, 1)

    def test_mock_with_stdout_stderr_returncode(self):
        mock_run = RunMocker()
        mock_run.register('wah',
                      stdout=['wah', 'wah wah', 'wah wah wah'],
                      stderr=['WAH', 'WAH WAH', 'WAH WAH WAH'],
                      returncode=520)

        p = mock_run(['wah'])
        self.eq(p.stdout.lines, ['wah', 'wah wah', 'wah wah wah'])
        self.eq(p.stderr.lines, ['WAH', 'WAH WAH', 'WAH WAH WAH'])
        self.eq(p.returncode, 520)

    def test_mock_with_stdout_stderr_returncode_side_effect(self):
        mock_run = RunMocker()
        mock_run.register('wah',
                      stdout=['wah1'],
                      stderr=['WAH1'],
                      returncode=42)
        mock_run.register('wah',
                      stdout=['wah2'],
                      stderr=['WAH2'],
                      returncode=43)

        p = mock_run(['wah'])
        self.eq(p.stdout.lines, ['wah1'])
        self.eq(p.stderr.lines, ['WAH1'])
        self.eq(p.returncode, 42)

        p = mock_run(['wah'])
        self.eq(p.stdout.lines, ['wah2'])
        self.eq(p.stderr.lines, ['WAH2'])
        self.eq(p.returncode, 43)

    def test_mock_side_effect(self):
        mock_run = RunMocker()

        def mock_ls_1st(proc, *args):
            self.eq(args, ('-a', '-l'))
            proc.stdout.writeline('file1')
        def mock_ls_3rd(proc, *args):
            self.eq(proc.cmd, ('ls', '-a', '-l', '--wah'))
            self.eq(args, ('-a', '-l', '--wah'))
            proc.stdout.writeline('file3')
        mock_run.register('ls', mock_ls_1st)
        mock_run.register('ls', stdout=['wah'], stderr=['WAH'], returncode=42)
        mock_run.register('ls', mock_ls_3rd)
        mock_run.register('ls', ValueError('ls is called too many times'))

        p = mock_run('ls -a -l'.split())
        self.eq(p.returncode, None)
        self.eq(p.stdout.lines, ['file1'])

        p = mock_run('ls -a -l'.split())
        self.eq(p.returncode, 42)
        self.eq(p.stdout.lines, ['wah'])
        self.eq(p.stderr.lines, ['WAH'])

        p = mock_run('ls -a -l --wah'.split())
        self.eq(p.returncode, None)
        self.eq(p.stdout.lines, ['file3'])

        with self.raises(ValueError):
            p = mock_run('ls -a -l --wah'.split())



import sys

from .lib_itertools import zip_longest, flatten

from .internal_utils import exporter
export, __all__ = exporter()


@export
def strwidth(s):
    import unicodedata
    from .lib_colors import decolor
    return sum((1 + (unicodedata.east_asian_width(c) in 'WF')) for c in decolor(s))


def lpad(text, padding):
    return text + padding


def rpad(text, padding):
    return padding + text


def just_elem(func):
    def wrapper(elem, width, fillchar):
        row, col, text = elem
        padding = (width - strwidth(text)) * fillchar(row=row, col=col, text=text)
        return func(text, padding)
    return wrapper


def just_generator(just_func, data, width, fillchar):
    for row, vector in enumerate(data):
        if isinstance(width, int):
            width = (width,) * len(vector)
        yield tuple(
                just_func((row, col, text), w, fillchar)
                for col, (text, w) in enumerate(zip_longest(vector, width[:len(vector)], fillvalues=('', 0)))
                )


def just(just_func, data, width, fillchar):
    if not callable(fillchar):
        _fillchar = fillchar
        fillchar = lambda row, col, text: _fillchar

    if isinstance(data, str):
        return just_func((0, 0, data), width, fillchar)

    if width:
        if isinstance(data, (tuple, list)):
            t = type(data)
        else:
            t = lambda x: x
        return t(just_generator(just_func, data, width, fillchar))

    maxwidth = []
    for vector in data:
        maxwidth = [
                max(w, strwidth(text))
                for text, w in zip_longest(vector, maxwidth, fillvalues=('', 0))
                ]

    return [
            tuple(
                just_func((row, col, text), w, fillchar)
                for col, (text, w) in enumerate(zip_longest(vector, maxwidth, fillvalues=('', 0)))
                )
            for row, vector in enumerate(data)
            ]


@export
def ljust(data, width=None, fillchar=' '):
    return just(just_elem(lpad), data, width, fillchar)


@export
def rjust(data, width=None, fillchar=' '):
    return just(just_elem(rpad), data, width, fillchar)


@export
class ThreadedSpinner:
    def __init__(self, *icon, delay=0.1):
        if not icon:
            self.icon_entry = '⠉⠛⠿⣿⠿⠛⠉⠙'
            self.icon_loop = '⠹⢸⣰⣤⣆⡇⠏⠛'
            self.icon_leave = '⣿'
        elif len(icon) == 1:
            self.icon_entry = tuple()
            self.icon_loop = icon
            self.icon_leave = '.'
        elif len(icon) == 2:
            self.icon_entry = icon[0]
            self.icon_loop = icon[1]
            self.icon_leave = '.'
        elif len(icon) == 3:
            self.icon_entry = icon[0]
            self.icon_loop = icon[1]
            self.icon_leave = icon[2]
        else:
            raise ValueError('Invalid value: ' + repr(icon))

        ok = True
        for name, icon in (('entry', self.icon_entry), ('loop', self.icon_loop), ('leave', self.icon_leave)):
            if isinstance(icon, str):
                ok = True
            elif isinstance(icon, (tuple, list)) and all(isinstance(c, str) for c in icon):
                ok = True
            else:
                raise ValueError('Invalid value of icon[{}]: {}'.format(name, icon))

        self.delay = delay
        self.is_end = False
        self.thread = None
        self._text = ''

        import itertools
        self.icon_iter = (
                itertools.chain(
                    self.icon_entry,
                    itertools.cycle(self.icon_loop)
                    ),
                iter(self.icon_leave)
                )
        self.icon_head = [None, None]

        import builtins
        self.print_function = builtins.print

    def __enter__(self):
        if self.thread:
            return self

        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end()

    @property
    def icon(self):
        idx = self.is_end
        if self.icon_head[idx] is None:
            self.icon_head[idx] = next(self.icon_iter[idx])
        return self.icon_head[idx]

    def text(self, *args):
        if not args:
            return self._text

        self._text = ' '.join(str(a) for a in args)
        if self.thread:
            self.refresh()

    def refresh(self):
        self.print_function('\r' + self.icon + '\033[K ' + self._text, end='')

    def animate(self):
        import time

        while not self.is_end:
            self.refresh()
            time.sleep(self.delay)
            self.icon_head[0] = next(self.icon_iter[0])

        try:
            while True:
                self.refresh()
                self.icon_head[1] = next(self.icon_iter[1])
                time.sleep(self.delay)
        except StopIteration:
            pass

        self.print_function()

    def start(self):
        if self.thread:
            return

        import threading
        self.thread = threading.Thread(target=self.animate)
        self.thread.daemon = True
        self.thread.start()

    def end(self, wait=True):
        self.is_end = True
        if wait:
            self.join()

    def join(self):
        self.thread.join()


def alt_if_none(A, B):
    if A is None:
        return B
    return A


class UserSelection:
    def __init__(self, options, accept_empty=None, abbr=None, sep=None, ignorecase=None):
        if not options:
            accept_empty = True
            abbr = False
            ignorecase = False

        self.accept_empty = alt_if_none(accept_empty, True)
        self.abbr = alt_if_none(abbr, True)
        self.ignorecase = alt_if_none(ignorecase, self.abbr)
        self.sep = alt_if_none(sep, ' / ')

        self.mapping = dict()
        self.options = [o for o in options]

        if self.options:
            if self.accept_empty:
                self.mapping[''] = self.options[0]

            for opt in self.options:
                for o in (opt,) + ((opt[0],) if self.abbr else tuple()):
                    self.mapping[o.lower() if self.ignorecase else o] = opt

        self.selected = None

    def select(self, o=''):
        if self.ignorecase:
            o = o.lower()

        if not self.options:
            self.selected = o
            return

        if o not in self.mapping:
            raise ValueError('Invalid option: ' + o)

        self.selected = o

    @property
    def prompt(self):
        if not self.options:
            return ''

        opts = [o for o in self.options]
        if self.accept_empty and self.ignorecase:
            opts[0] = opts[0].capitalize()

        if self.abbr:
            return ' [' + self.sep.join('({}){}'.format(o[0], o[1:]) for o in opts) + ']'
        else:
            return ' [' + self.sep.join(opts) + ']'

    def __eq__(self, other):
        if self.ignorecase and other is not None:
            other = other.lower()

        if self.selected == other:
            return True

        if self.selected in self.mapping:
            return self.mapping[self.selected] == self.mapping.get(other)

        return False

    def __str__(self):
        return str(self.selected)

    def __repr__(self):
        return '<warawara.tui.UserSelection selected=[{}]>'.format(self.selected)


class HijackStdio:
    def __init__(self, replace_with='/dev/tty'):
        self.replace_with = replace_with

    def __enter__(self):
        self.stdin_backup = sys.stdin
        self.stdout_backup = sys.stdout
        self.stderr_backup = sys.stderr

        sys.stdin = open(self.replace_with)
        sys.stdout = open(self.replace_with, 'w')
        sys.stderr = open(self.replace_with, 'w')

    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdin.close()
        sys.stdout.close()
        sys.stderr.close()

        sys.stdin = self.stdin_backup
        sys.stdout = self.stdout_backup
        sys.stderr = self.stderr_backup


class ExceptionSuppressor:
    def __init__(self, *exc_group):
        if isinstance(exc_group[0], tuple):
            self.exc_group = exc_group[0]
        else:
            self.exc_group = exc_group

    def __enter__(self, *exc_group):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type in (EOFError, KeyboardInterrupt):
            print()
        return exc_type in self.exc_group


@export
def prompt(question, options=tuple(),
           accept_empty=True,
           abbr=True,
           ignorecase=None,
           sep=' / ',
           suppress=(EOFError, KeyboardInterrupt)):

    if isinstance(options, str) and ' ' in options:
        options = options.split()

    user_selection = UserSelection(options, accept_empty=accept_empty, abbr=abbr, sep=sep, ignorecase=ignorecase)

    with HijackStdio():
        with ExceptionSuppressor(suppress):
            while user_selection.selected is None:
                print((question + (user_selection.prompt)), end=' ')

                import contextlib
                with contextlib.suppress(ValueError):
                    i = input().strip()
                    user_selection.select(i)

    return user_selection


class Key:
    def __init__(self, seq, *aliases):
        if isinstance(seq, str):
            seq = seq.encode('utf8')

        if not isinstance(seq, bytes):
            raise TypeError('seq should be in type bytes, not {}'.format(type(seq)))

        if not all(isinstance(a, str) for a in aliases):
            raise TypeError('Aliases should be in type str')

        self.seq = seq
        self.aliases = []
        for name in aliases:
            self.nameit(str(name))

    def __hash__(self):
        return hash(self.seq)

    def __repr__(self):
        fmt = type(self).__name__ + '({})'
        if self.aliases:
            return fmt.format(self.aliases[0])
        try:
            return fmt.format(repr(self.seq.decode('utf8')))
        except UnicodeError:
            return fmt.format(repr(self.seq))

    def nameit(self, name):
        if name not in self.aliases:
            self.aliases.append(name)

    def __eq__(self, other):
        if type(self) == type(other):
            return self.seq == other.seq
        elif isinstance(other, bytes) and self.seq == other:
            return True
        elif isinstance(other, str) and self.seq == other.encode('utf8'):
            return True
        else:
            return other in self.aliases


KEY_ESCAPE = Key(b'\033', 'esc', 'escape')
KEY_BACKSPACE = Key(b'\x7f', 'backspace')
KEY_TAB = Key(b'\t', 'tab', 'ctrl-i', 'ctrl+i', '^I')
KEY_ENTER = Key(b'\r', 'enter', 'ctrl-m', 'ctrl+m', '^M')
KEY_SPACE = Key(b' ', 'space')

KEY_UP = Key(b'\033[A', 'up')
KEY_DOWN = Key(b'\033[B', 'down')
KEY_RIGHT = Key(b'\033[C', 'right')
KEY_LEFT = Key(b'\033[D', 'left')

KEY_HOME = Key(b'\033[1~', 'home')
KEY_END = Key(b'\033[4~', 'end')
KEY_PGUP = Key(b'\033[5~', 'pgup', 'pageup')
KEY_PGDN = Key(b'\033[6~', 'pgdn', 'pagedown')

KEY_F1 = Key(b'\033OP', 'F1')
KEY_F2 = Key(b'\033OQ', 'F2')
KEY_F3 = Key(b'\033OR', 'F3')
KEY_F4 = Key(b'\033OS', 'F4')
KEY_F5 = Key(b'\033[15~', 'F5')
KEY_F6 = Key(b'\033[17~', 'F6')
KEY_F7 = Key(b'\033[18~', 'F7')
KEY_F8 = Key(b'\033[19~', 'F8')
KEY_F9 = Key(b'\033[20~', 'F9')
KEY_F10 = Key(b'\033[21~', 'F10')
KEY_F11 = Key(b'\033[23~', 'F11')
KEY_F12 = Key(b'\033[24~', 'F12')

def _register_ctrl_n_keys():
    for c in 'abcdefghjklnopqrstuvwxyz':
        C = c.upper()
        idx = ord(c) - ord('a') + 1
        aliases = ('ctrl-' + c, 'ctrl+' + c, '^' + C)
        globals()['KEY_CTRL_' + C] = Key(chr(idx), *aliases)

_register_ctrl_n_keys()
del _register_ctrl_n_keys


def _export_all_keys():
    for key in globals().keys():
        if key.startswith('KEY_'):
            export(key)

_export_all_keys()
del _export_all_keys


key_table = {}
key_table_reverse = {}

def _init_key_table():
    for k, v in globals().items():
        if not k.startswith('KEY_'):
            continue
        key_table[v.seq] = v

        for alias in v.aliases:
            key_table_reverse[alias] = v

_init_key_table()
del _init_key_table


@export
def register_key(seq, *aliases):
    if isinstance(seq, Key):
        new_key = seq
        seq = new_key.seq
        aliases = new_key.aliases + list(aliases)

    elif isinstance(seq, str):
        seq = seq.encode('utf8')

    if not seq:
        raise ValueError('huh?')

    if seq not in key_table:
        key_table[seq] = Key(seq, *aliases)
        return key_table[seq]

    key = key_table[seq]
    for name in aliases:
        key.nameit(name)

    return key


@export
def deregister_key(seq):
    if isinstance(seq, Key):
        seq = seq.seq
    elif isinstance(seq, str):
        seq = seq.encode('utf8')
    return key_table.pop(seq, None)


@export
def getch(timeout=None, encoding='utf8'):
    import termios, tty
    import os
    import select

    fd = sys.stdin.fileno()
    orig_term_attr = termios.tcgetattr(fd)
    when = termios.TCSADRAIN

    def has_data(t=0):
        return select.select([fd], [], [], t)[0]

    def read_one_byte():
        return os.read(sys.stdin.fileno(), 1)

    try:
        tty.setraw(fd, when=when)

        # Wait for input until timeout
        if not has_data(timeout):
            return None

        acc = b''
        candidate_matches = set(key_table.keys())
        while True:
            acc += read_one_byte()

            if not has_data():
                break

            # Still have chance to match in key table
            if candidate_matches:
                # eliminate potential matches
                candidate_matches = set(key_seq for key_seq in candidate_matches if key_seq.startswith(acc))

                # Perfect match, return
                if candidate_matches == {acc}:
                    break

                # multiple prefix matchs: collect more byte
                if candidate_matches:
                    continue

            # Input sequence does not match anything in key table
            # Collect enough bytes to decode at least one unicode char
            try:
                acc.decode(encoding)
                break
            except UnicodeError:
                continue

        if acc in key_table:
            return key_table[acc]

        try:
            return acc.decode(encoding)
        except UnicodeError:
            return acc

    finally:
        termios.tcsetattr(fd, when, orig_term_attr)



__version__ = '0.2.0'


def check_python_version():
    import sys
    assert sys.version_info.major, sys.version_info.minor >= (3, 7)

check_python_version()
del check_python_version


def load_internal_modules(prefix, post_load_hook):
    import importlib
    import pkgutil
    scope_prefix = __name__ + '.'
    for mod_info in pkgutil.iter_modules(__path__, scope_prefix):
        mod_name = mod_info.name[len(scope_prefix):]
        if not mod_name.startswith(prefix):
            continue

        module = importlib.import_module(mod_info.name)
        post_load_hook(module, mod_name)


def post_load_lib(module, mod_name):
    ext_name = mod_name[4:]

    if '__all__' in module.__dict__:
        attrs = module.__dict__['__all__']
    else:
        attrs = [x for x in module.__dict__ if not x.startswith('_')]

    # Register module into package namespace with external name
    globals()[ext_name] = module
    globals().update({attr: getattr(module, attr) for attr in attrs})

    # Delete old name from package namespace
    del globals()[mod_name]


def post_load_bin(module, mod_name):
    ext_name = mod_name[4:]

    # Register module into warawara.bin with external name
    setattr(bin, ext_name, module)

    # Delete old name from package namespace
    del globals()[mod_name]


from . import bin

load_internal_modules('lib_', post_load_lib)
load_internal_modules('bin_', post_load_bin)

del post_load_lib
del post_load_bin
del load_internal_modules



import argparse
import shlex
import sys


from . import lib_subproc as subproc
from . import lib_colors as paints


def print_cmd(cmd):

    def color_token(arg):
        token = shlex.quote(arg)
        color = paints.nocolor
        if token.startswith(('"', "'")):
            color = paints.orange
        elif token.startswith('-'):
            color = paints.cyan

        return color(token)

    print(paints.magenta('$'), ' '.join(
        color_token(arg) for arg in cmd
        ))


def notify(title, lines):
    def quote(text):
        return '"' + text.replace('"', r'\"') + '"'

    args = ['display', 'notification', quote(r'\n'.join(lines))]

    if title is not None:
        args += ['with', 'title', quote(title)]

    cmd = ['osascript', '-e', ' '.join(args)]
    print_cmd(cmd)
    exit(subproc.run(cmd).returncode)


def main():
    prog = sys.argv[0]
    argv = sys.argv[1:]

    parser = argparse.ArgumentParser(description='ntfy', prog='ntfy')
    parser.add_argument('-t', '--title', help='Notification title')
    parser.add_argument('lines', nargs='*', help='Notification text', default=[])

    args = parser.parse_args(argv)

    notify(args.title, args.lines)



import itertools

from .internal_utils import exporter
export, __all__ = exporter()


@export
def is_iterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False


@export
def unwrap_one(obj):
    try:
        while True:
            if len(obj) == 1 and iter(obj[0]) and not isinstance(obj[0], str):
                obj = obj[0]
            else:
                return obj
    except TypeError:
        pass

    return obj


@export
def unwrap(obj=None):
    try:
        while True:
            if isinstance(obj, str):
                return obj

            if len(obj) == 1:
                obj = obj[0]
                continue

            return obj
    except TypeError:
        pass

    return obj


@export
def flatten(tree):
    if not is_iterable(tree) or isinstance(tree, str):
        return tree

    wrapper_type = type(tree)
    return wrapper_type(itertools.chain.from_iterable(
        flatten(i) if is_iterable(i) and not isinstance(i, str) else [i]
        for i in tree
        ))


@export
def lookahead(iterable):
    it = iter(iterable)
    lookahead = next(it)

    for val in it:
        yield lookahead, False
        lookahead = val

    yield lookahead, True


@export
def zip_longest(*iterables, fillvalues=None):
    if not isinstance(fillvalues, (tuple, list)):
        fillvalues = (fillvalues,) * len(iterables)

    iterators = list(map(iter, iterables))

    while True:
        values = []
        cont = False
        for idx, iterator in enumerate(iterators):
            try:
                value = next(iterator)
                cont = True
            except:
                value = fillvalues[idx]

            values.append(value)

        if not cont:
            break

        yield tuple(values)



import argparse
import datetime
import sys
import threading
import time

from . import lib_subproc as subproc


def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def worker(streams, cmd, delay, stop_signal):
    # Get initial content from command for excluding it
    p = subproc.run(cmd)
    if p.returncode:
        return

    old_lines = p.stdout.lines
    sponged_lines = []

    while True:
        if stop_signal.is_set():
            break

        p = subproc.run(cmd)

        if p.returncode:
            break

        new_lines = p.stdout.lines

        if old_lines != new_lines:
            for line in new_lines:
                sponged_lines.append(line)

                # Print timestamp for reference so you know what's going on
                if not sys.stdout.isatty() or True:
                    print_err('[' + str(datetime.datetime.now()) + ']', line)

            old_lines = new_lines

            subproc.run(['ntfy', '-t', 'Copied', '{} lines'.format(len(sponged_lines))])

        if cmd[0] != 'sleep':
            time.sleep(delay)

    streams[1].writelines(sponged_lines)


def main():
    try:
        parser = argparse.ArgumentParser(prog='sponge',
                description='sponge',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument('-d', '--delay', default=0.2, type=float, help='Delay in seconds')
        parser.add_argument('command', nargs='*', default=None, help='Command to run')

        args = parser.parse_args()

        if not args.command:
            sponged_lines = [line.rstrip() for line in sys.stdin.readlines()]
            for line in sponged_lines:
                print(line)
            sys.exit()

        stop_signal = threading.Event()

        cmd = subproc.run([worker, args.command, args.delay, stop_signal], stderr=None, wait=False)

        for line in sys.stdin:
            print('[ignored]', line.rstrip())
        stop_signal.set()

        cmd.wait()

        for line in cmd.stdout:
            print(line)

    except KeyboardInterrupt:
        print_err('KeyboardInterrupt')



from .lib_itertools import unwrap_one

from .internal_utils import exporter
export, __all__ = exporter()


@export
def is_uint8(i):
    return isinstance(i, int) and not isinstance(i, bool) and 0 <= i < 256


@export
def sgn(i):
    return (i > 0) - (i < 0)


@export
def lerp(a, b, t):
    if t == 0:
        return a
    if t == 1:
        return b
    return a + t * (b - a)


@export
def clamp(A, x, B):
    import builtins
    min, max = builtins.min(A, B), builtins.max(A, B)
    return builtins.max(min, builtins.min(x, max))


@export
class vector:
    def __init__(self, *args):
        args = unwrap_one(args)

        if isinstance(args, vector):
            self.data = list(args.data)
        else:
            self.data = list(args)

        for i in self.data:
            if not isinstance(i, (int, float, complex)):
                raise ValueError('Invalid value: {}'.format(repr(tuple(self.data))))

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __eq__(self, other):
        if not isinstance(other, (vector, tuple, list)):
            raise TypeError('Cannot compare with type {}'.format(repr(type(other))))
        return tuple(self.data) == tuple(other)

    def __repr__(self):
        return '(' + ', '.join(map(str, self.data)) + ')'

    def __add__(self, other):
        if isinstance(other, (int, float)):
            return vector(i + other for i in self)
        if len(self) != len(other):
            raise ValueError('Cannot operate on vector(len={}) and {}'.format(len(self), other))
        return vector(map(lambda x: x[0] + x[1], zip(self, other)))

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            return vector(i - other for i in self)
        if len(self) != len(other):
            raise ValueError('Cannot operate on vector(len={}) and {}'.format(len(self), other))
        return vector(map(lambda x: x[0] - x[1], zip(self, other)))

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return vector(i * other for i in self)
        if len(self) != len(other):
            raise ValueError('Cannot operate on vector(len={}) and {}'.format(len(self), other))
        return vector(map(lambda x: x[0] * x[1], zip(self, other)))

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return vector(i / other for i in self)
        raise TypeError('Cannot operate on vector(len={}) and {}'.format(len(self), other))

    def __floordiv__(self, other):
        if isinstance(other, (int, float)):
            return vector(i // other for i in self)
        raise TypeError('Cannot operate on vector(len={}) and {}'.format(len(self), other))

    def map(self, func):
        return vector(func(i) for i in self)


@export
def interval(a, b, close=True):
    direction = sgn(b - a)
    if direction == 0:
        return [a] if close else []

    ret = range(a, b + direction, direction)
    if close:
        return list(ret)
    return list(ret[1:-1])


@export
def resample(samples, N):
    if N is None:
        return samples

    n = len(samples)

    if N == n:
        return samples

    if N < n:
        # Averaging skipped samples into N-1 gaps
        skip_count = n - N
        gap_count = N - 1

        probe = 0
        dup, rem = divmod(skip_count, gap_count)

        ret = [samples[0]]
        for i in range(gap_count):
            probe += 1 + dup + (i < rem)
            ret.append(samples[probe])

    if N > n:
        # Duplicate samples to match N
        ret = []
        dup, rem = divmod(N, n)
        for i in range(n):
            for d in range(dup + (i < rem)):
                ret.append(samples[i])

    return tuple(ret)



import queue

import functools
import threading
import unittest.mock

from collections import namedtuple

from .lib_test_utils import *

from warawara import *


class TestTypesettingUtils(TestCase):
    def test_strwidth(self):
        self.eq(strwidth('test'), 4)
        self.eq(strwidth(orange('test')), 4)
        self.eq(strwidth('哇嗚'), 4)

    def test_ljust_str(self):
        self.eq(ljust('test', 10), 'test      ')
        self.eq(rjust('test', 10), '      test')

        padding = ' ' * 6
        self.eq(ljust(orange('test'), 10), orange('test') + padding)
        self.eq(rjust(orange('test'), 10), padding + orange('test'))

        padding = '#' * 6
        self.eq(ljust(orange('test'), 10, '#'), orange('test') + padding)
        self.eq(rjust(orange('test'), 10, '#'), padding + orange('test'))

    def test_just_rect(self):
        data = [
                ('column1', 'col2'),
                ('word1', 'word2'),
                ('word3', 'word4 long words'),
                ]

        self.eq(ljust(data), [
            ('column1', 'col2            '),
            ('word1  ', 'word2           '),
            ('word3  ', 'word4 long words'),
            ])

        self.eq(rjust(data), [
            ('column1', '            col2'),
            ('  word1', '           word2'),
            ('  word3', 'word4 long words'),
            ])

    def test_just_with_fillchar(self):
        data = [
                ('column1', 'col2'),
                ('word1', 'word2'),
                ('word3', 'word4 long words'),
                ]

        self.eq(ljust(data, fillchar='#'), [
            ('column1', 'col2############'),
            ('word1##', 'word2###########'),
            ('word3##', 'word4 long words'),
            ])

    def test_just_with_fillchar_func(self):
        data = [
                ('up left',   'up',   'up right'),
                ('left',      '',     'right'),
                ('down left', 'down', 'down r'),
                ]

        def fillchar(row, col, text):
            if row + col == 2:
                return '%'
            if text == 'right':
                return '$'
            return '#' if (row % 2) else '@'

        self.eq(ljust(data, fillchar=fillchar, width=10), [
            ('up left@@@', 'up@@@@@@@@', 'up right%%'),
            ('left######', '%%%%%%%%%%', 'right$$$$$'),
            ('down left%', 'down@@@@@@', 'down r@@@@'),
            ])

        self.eq(rjust(data, fillchar=fillchar, width=10), [
            ('@@@up left', '@@@@@@@@up', '%%up right'),
            ('######left', '%%%%%%%%%%', '$$$$$right'),
            ('%down left', '@@@@@@down', '@@@@down r'),
            ])

    def test_just_with_width(self):
        data = [
                ('column1', 'col2'),
                ('word1', 'word2'),
                ('word3', 'word4 long words'),
                ]

        self.eq(ljust(data, width=20), [
            ('column1             ', 'col2                '),
            ('word1               ', 'word2               '),
            ('word3               ', 'word4 long words    '),
            ])

        self.eq(ljust(data, width=(10, 20)), [
            ('column1   ', 'col2                '),
            ('word1     ', 'word2               '),
            ('word3     ', 'word4 long words    '),
            ])

    def test_just_with_generator(self):
        data = [
                ('column1', 'col2'),
                ('word1', 'word2'),
                ('word3', 'word4 long words'),
                ]

        ret = ljust((vector for vector in data), width=(10, 20))
        self.false(isinstance(ret, (tuple, list)))

        self.eq(list(ret), [
            ('column1   ', 'col2                '),
            ('word1     ', 'word2               '),
            ('word3     ', 'word4 long words    '),
            ])

    def test_just_rect_lack_columns(self):
        self.eq(
                ljust([
                    ('column1', 'col2'),
                    ('word1',),
                    tuple(),
                    ('', 'multiple words'),
                    tuple(),
                    ]),
                [
                    ('column1', 'col2          '),
                    ('word1  ', '              '),
                    ('       ', '              '),
                    ('       ', 'multiple words'),
                    ('       ', '              '),
                    ])

    def test_just_rect_more_columns(self):
        self.eq(
                ljust([
                    ('column1', 'col2'),
                    tuple(),
                    ('word1', 'word2', 'word4'),
                    ('word3', 'multiple words'),
                    ]),
                [
                    ('column1', 'col2          ', '     '),
                    ('       ', '              ', '     '),
                    ('word1  ', 'word2         ', 'word4'),
                    ('word3  ', 'multiple words', '     '),
                    ])


def queue_to_list(Q):
    ret = []
    while not Q.empty():
        ret.append(Q.get())
    return ret


class TestThreadedSpinner(TestCase):
    Event = namedtuple('Event',
                       ('timestamp', 'tag', 'args', 'callback'),
                       defaults=(None, None, None, None))

    def setUp(self):
        self.sys_time = 0
        self.behavior_queue = queue.Queue()
        self.events_upon_sleep = queue.Queue()

    def mock_print(self, *args, **kwargs):
        if not args and not kwargs:
            self.behavior_queue.put((self.sys_time, 'print', None))
        else:
            self.behavior_queue.put((
                self.sys_time, 'print',
                tuple(' '.join(args).lstrip('\r').split('\x1b[K ')),
                ))

    def mock_sleep(self, secs):
        self.behavior_queue.put((self.sys_time, 'sleep', secs))

        if not self.events_upon_sleep.empty():
            callback = self.events_upon_sleep.get()
            if callable(callback):
                callback()

            self.events_upon_sleep.task_done()

        self.sys_time += secs

    def test_default_values(self):
        self.patch('builtins.print', RuntimeError('Should not print() at all'))
        self.patch('time.sleep', RuntimeError('Should not sleep() at all'))

        spinner = ThreadedSpinner()
        self.eq(spinner.delay, 0.1)
        self.eq(spinner.icon_entry, '⠉⠛⠿⣿⠿⠛⠉⠙')
        self.eq(spinner.icon_loop, '⠹⢸⣰⣤⣆⡇⠏⠛')
        self.eq(spinner.icon_leave, '⣿')
        self.eq(spinner.text(), '')
        spinner.text('wah')
        self.eq(spinner.text(), 'wah')

    def test_icon_set_loop(self):
        spinner = ThreadedSpinner('LOOP')
        self.eq(spinner.icon_entry, tuple())
        self.eq(spinner.icon_loop, ('LOOP',))
        self.eq(spinner.icon_leave, '.')

    def test_icon_set_entry_loop(self):
        spinner = ThreadedSpinner('ENTRY', 'LOOP')
        self.eq(spinner.icon_entry, 'ENTRY')
        self.eq(spinner.icon_loop, 'LOOP')
        self.eq(spinner.icon_leave, '.')

    def test_icon_set_entry_loop_leave(self):
        spinner = ThreadedSpinner('ENTRY', 'LOOP', 'LEAVE')
        self.eq(spinner.icon_entry, 'ENTRY')
        self.eq(spinner.icon_loop, 'LOOP')
        self.eq(spinner.icon_leave, 'LEAVE')

    def test_icon_set_invalid(self):
        with self.raises(ValueError):
            spinner = ThreadedSpinner('ENTRY', 'LOOP', 'LEAVE', 'WHAT')

        with self.raises(ValueError):
            spinner = ThreadedSpinner(True)

    def test_context_manager(self):
        spinner = ThreadedSpinner()
        spinner.print_function = lambda *args, **kwarags: None
        with spinner:
            with spinner:
                spinner.start()

    def test_run(self):
        self.patch('time.sleep', self.mock_sleep)
        Event = self.__class__.Event

        delay = 1
        spinner = ThreadedSpinner('ENTRY', 'LOOP', 'OUT', delay=delay)
        spinner.print_function = self.mock_print

        event_list = [
                Event( 0, 'print', ('E', 'meow')),
                Event( 0, 'sleep', delay),
                Event( 1, 'print', ('N', 'meow')),
                Event( 1, 'sleep', delay),
                Event( 2, 'print', ('T', 'meow')),
                Event( 2, 'sleep', delay),
                Event( 3, 'print', ('R', 'meow')),
                Event( 3, 'sleep', delay),
                Event( 4, 'print', ('Y', 'meow')),
                Event( 4, 'sleep', delay),
                Event( 5, 'print', ('L', 'meow')),
                Event( 5, 'sleep', delay),
                Event( 6, 'print', ('O', 'meow')),
                Event( 6, 'sleep', delay),
                Event( 7, 'print', ('O', 'meow')),
                Event( 7, 'sleep', delay),
                Event( 8, 'print', ('P', 'meow')),
                Event( 8, 'sleep', delay),
                Event( 9, 'print', ('L', 'meow')),
                Event( 9, 'sleep', delay, functools.partial(spinner.text, 'woof')),
                Event( 9, 'print', ('L', 'woof')),
                Event(10, 'print', ('O', 'woof')),
                Event(10, 'sleep', delay),
                Event(11, 'print', ('O', 'woof')),
                Event(11, 'sleep', delay),
                Event(12, 'print', ('P', 'woof')),
                Event(12, 'sleep', delay),
                Event(13, 'print', ('L', 'woof')),
                Event(13, 'sleep', delay),
                Event(14, 'print', ('O', 'woof')),
                Event(14, 'sleep', delay),
                Event(15, 'print', ('O', 'woof')),
                Event(15, 'sleep', delay),
                Event(16, 'print', ('P', 'woof')),
                Event(16, 'sleep', delay, functools.partial(spinner.end, wait=False)),
                Event(17, 'print', ('O', 'woof')),
                Event(17, 'sleep', delay),
                Event(18, 'print', ('U', 'woof')),
                Event(18, 'sleep', delay),
                Event(19, 'print', ('T', 'woof')),
                Event(19, 'print'),
                ]

        for event in filter(lambda e: e.tag == 'sleep', event_list):
            self.events_upon_sleep.put(event.callback)

        spinner.text('meow')
        spinner.start()
        spinner.join()

        from itertools import zip_longest
        for e, behavior in zip_longest(event_list, queue_to_list(self.behavior_queue)):
            expected = (e.timestamp, e.tag, e.args)
            self.eq(expected, behavior)


class TestPromotAskUser(TestCase):
    def setUp(self):
        self.input_queue = None
        self.print_queue = queue.Queue()

        self.patch('builtins.print', self.mock_print)
        self.patch('builtins.input', self.mock_input)

        self.mock_open = unittest.mock.mock_open()
        self.patch('builtins.open', self.mock_open)
        self.assert_called_open = True

    def tearDown(self):
        if self.assert_called_open:
            self.mock_open.assert_has_calls([
                    unittest.mock.call('/dev/tty'),
                    unittest.mock.call('/dev/tty', 'w'),
                    unittest.mock.call('/dev/tty', 'w'),
                    ])

            handle = self.mock_open()
            handle.close.assert_has_calls([
                    unittest.mock.call(),
                    unittest.mock.call(),
                    unittest.mock.call(),
                ])
        else:
            self.mock_open.assert_not_called()

    def set_input(self, *lines):
        self.input_queue = queue.Queue()
        for line in lines:
            self.input_queue.put(line)

    def mock_print(self, *args, **kwargs):
        self.print_queue.put(' '.join(args) + kwargs.get('end', '\n'))

    def mock_input(self, prompt=None):
        if prompt:
            self.print_queue.put(prompt)

        if self.input_queue.empty():
            self.fail('Expected more test input')

        s = self.input_queue.get()
        if isinstance(s, BaseException):
            raise s
        return s + '\n'

    def expect_output(self, *args):
        self.eq(queue_to_list(self.print_queue), list(args))

    def test_empty(self):
        with self.raises(TypeError):
            s = prompt()

        self.assert_called_open = False

    def test_continue(self):
        self.set_input('wah')
        yn = prompt('Input anything to continue>')
        self.expect_output('Input anything to continue> ')

        repr(yn)
        self.eq(yn.selected, 'wah')
        self.eq(str(yn), 'wah')
        self.eq(yn, 'wah')
        self.ne(yn, 'WAH')

    def test_coffee_or_tea(self):
        self.set_input('')
        yn = prompt('Coffee or tea?', 'coffee tea')
        self.expect_output('Coffee or tea? [(C)offee / (t)ea] ')

        self.eq(yn.ignorecase, True)
        self.eq(yn.selected, '')
        self.eq(yn, '')
        self.eq(yn, 'coffee')
        self.eq(yn, 'Coffee')
        self.eq(yn, 'COFFEE')
        self.ne(yn, 'tea')

    def test_coffee_or_tea_yes(self):
        self.set_input(
                'what',
                'WHAT?',
                'tea',
                )
        yn = prompt('Coffee or tea?', 'coffee tea both')
        self.expect_output(
                'Coffee or tea? [(C)offee / (t)ea / (b)oth] ',
                'Coffee or tea? [(C)offee / (t)ea / (b)oth] ',
                'Coffee or tea? [(C)offee / (t)ea / (b)oth] ',
                )

        self.eq(yn.selected, 'tea')
        self.eq(yn, 'tea')
        self.ne(yn, 'coffee')

    def test_eoferror(self):
        self.set_input(EOFError())
        yn = prompt('Coffee or tea?', 'coffee tea')
        self.expect_output(
                'Coffee or tea? [(C)offee / (t)ea] ',
                '\n',
                )

        self.eq(yn.selected, None)
        self.eq(yn, None)
        self.ne(yn, 'coffee')
        self.ne(yn, 'tea')
        self.ne(yn, 'water')
        self.ne(yn, 'both')

    def test_keyboardinterrupt(self):
        self.set_input(KeyboardInterrupt())
        yn = prompt('Coffee or tea?', 'coffee tea')
        self.expect_output(
                'Coffee or tea? [(C)offee / (t)ea] ',
                '\n',
                )

        self.eq(yn.selected, None)
        self.eq(yn, None)
        self.ne(yn, 'coffee')
        self.ne(yn, 'tea')
        self.ne(yn, 'water')
        self.ne(yn, 'both')

    def test_suppress(self):
        self.set_input(RuntimeError('wah'), TimeoutError('waaaaah'))
        yn = prompt('Question', suppress=RuntimeError)
        self.eq(yn, None)

        with self.raises(TimeoutError):
            yn = prompt('Question', suppress=RuntimeError)
        self.eq(yn, None)

    def test_sep(self):
        self.set_input('')
        yn = prompt('Coffee or tea?', 'coffee tea', sep='|')
        self.expect_output('Coffee or tea? [(C)offee|(t)ea] ')

    def test_noignorecase(self):
        self.set_input('tea')
        yn = prompt('Coffee or tea?', 'coffee tea', ignorecase=False)
        self.expect_output('Coffee or tea? [(c)offee / (t)ea] ')

        self.eq(yn, 'tea')
        self.ne(yn, 'TEA')

        self.set_input('coFFee', 'coffEE', 'coffee')
        yn = prompt('Coffee or tea?', 'coffee tea', ignorecase=False)
        self.expect_output(
                'Coffee or tea? [(c)offee / (t)ea] ',
                'Coffee or tea? [(c)offee / (t)ea] ',
                'Coffee or tea? [(c)offee / (t)ea] ',
                )

        self.ne(yn, 'coFFee')
        self.ne(yn, 'coffEE')
        self.eq(yn, 'coffee')

    def test_noabbr(self):
        self.set_input('t', 'tea')
        yn = prompt('Coffee or tea?', 'coffee tea', abbr=False)
        self.expect_output(
                'Coffee or tea? [coffee / tea] ',
                'Coffee or tea? [coffee / tea] ',
                )

        self.ne(yn, 't')
        self.eq(yn, 'tea')
        self.ne(yn, 'TEA')

    def test_noaccept_empty(self):
        self.set_input('', 'c')
        yn = prompt('Coffee or tea?', 'coffee tea', accept_empty=False)
        self.expect_output(
                'Coffee or tea? [(c)offee / (t)ea] ',
                'Coffee or tea? [(c)offee / (t)ea] ',
                )

        self.eq(yn.ignorecase, True)
        self.eq(yn.selected, 'c')
        self.eq(yn, 'c')
        self.eq(yn, 'coffee')
        self.eq(yn, 'Coffee')
        self.eq(yn, 'COFFEE')
        self.ne(yn, 'tea')


class TestKey(TestCase):
    def test_builtin_key(self):
        self.eq(KEY_ESCAPE, b'\033')
        self.eq(KEY_ESCAPE, '\033')
        self.eq(KEY_ESCAPE, 'esc')
        self.eq(KEY_ESCAPE, 'escape')

        self.eq(KEY_BACKSPACE, b'\x7f')
        self.eq(KEY_BACKSPACE, 'backspace')

        self.eq(KEY_TAB, b'\t')
        self.eq(KEY_TAB, 'tab')
        self.eq(KEY_TAB, 'ctrl-i')
        self.eq(KEY_TAB, 'ctrl+i')
        self.eq(KEY_TAB, '^I')

        self.eq(KEY_ENTER, b'\r')
        self.eq(KEY_ENTER, '\r')
        self.eq(KEY_ENTER, 'enter')
        self.eq(KEY_ENTER, 'ctrl-m')
        self.eq(KEY_ENTER, 'ctrl+m')
        self.eq(KEY_ENTER, '^M')

        self.eq(KEY_SPACE, b' ')
        self.eq(KEY_SPACE, ' ')
        self.eq(KEY_SPACE, 'space')

        self.eq(KEY_UP, b'\033[A')
        self.eq(KEY_UP, '\033[A')
        self.eq(KEY_UP, 'up')

        self.eq(KEY_DOWN, b'\033[B')
        self.eq(KEY_DOWN, '\033[B')
        self.eq(KEY_DOWN, 'down')

        self.eq(KEY_RIGHT, b'\033[C')
        self.eq(KEY_RIGHT, '\033[C')
        self.eq(KEY_RIGHT, 'right')

        self.eq(KEY_LEFT, b'\033[D')
        self.eq(KEY_LEFT, '\033[D')
        self.eq(KEY_LEFT, 'left')

        self.eq(KEY_HOME, b'\033[1~')
        self.eq(KEY_HOME, '\033[1~')
        self.eq(KEY_HOME, 'home')

        self.eq(KEY_END, b'\033[4~')
        self.eq(KEY_END, '\033[4~')
        self.eq(KEY_END, 'end')

        self.eq(KEY_PGUP, b'\033[5~')
        self.eq(KEY_PGUP, '\033[5~')
        self.eq(KEY_PGUP, 'pgup')
        self.eq(KEY_PGUP, 'pageup')

        self.eq(KEY_PGDN, b'\033[6~')
        self.eq(KEY_PGDN, '\033[6~')
        self.eq(KEY_PGDN, 'pgdn')
        self.eq(KEY_PGDN, 'pagedown')

        self.eq(KEY_F1, b'\033OP')
        self.eq(KEY_F1, 'F1')

        self.eq(KEY_F2, b'\033OQ')
        self.eq(KEY_F2, 'F2')

        self.eq(KEY_F3, b'\033OR')
        self.eq(KEY_F3, 'F3')

        self.eq(KEY_F4, b'\033OS')
        self.eq(KEY_F4, 'F4')

        self.eq(KEY_F5, b'\033[15~')
        self.eq(KEY_F5, 'F5')

        self.eq(KEY_F6, b'\033[17~')
        self.eq(KEY_F6, 'F6')

        self.eq(KEY_F7, b'\033[18~')
        self.eq(KEY_F7, 'F7')

        self.eq(KEY_F8, b'\033[19~')
        self.eq(KEY_F8, 'F8')

        self.eq(KEY_F9, b'\033[20~')
        self.eq(KEY_F9, 'F9')

        self.eq(KEY_F10, b'\033[21~')
        self.eq(KEY_F10, 'F10')

        self.eq(KEY_F11, b'\033[23~')
        self.eq(KEY_F11, 'F11')

        self.eq(KEY_F12, b'\033[24~')
        self.eq(KEY_F12, 'F12')

        for c in 'abcdefghjklnopqrstuvwxyz':
            key = globals()['KEY_CTRL_' + c.upper()]
            self.eq(key, chr(ord(c) - ord('a') + 1))
            self.eq(key, 'ctrl-' + c)
            self.eq(key, 'ctrl+' + c)
            self.eq(key, '^' + c.upper())

    def test_key_invalid_seq_and_alias(self):
        Key = type(KEY_UP)
        with self.raises(TypeError):
            Key(['wah'], 'WAH')

        with self.raises(TypeError):
            Key('wah', ['WAH'])

    def test_key_hash(self):
        self.eq(hash(KEY_UP), hash(KEY_UP.seq))

    def test_key_nameit(self):
        Key = type(KEY_UP)
        TEST_KEY = Key('test_key')
        self.ne(TEST_KEY, 'wah')
        TEST_KEY.nameit('wah')
        self.eq(TEST_KEY, 'wah')
        TEST_KEY.nameit('wah')
        self.eq(TEST_KEY, 'wah')

    def test_key_repr(self):
        self.eq(repr(KEY_UP), 'Key(up)')

        Key = type(KEY_UP)
        new_key = Key('測')
        self.eq(repr(new_key), r"Key('測')")

        seq = '測'.encode('utf8')[:-2]
        new_key2 = Key(seq)
        self.eq(repr(new_key2), r'Key(' + repr(seq) + ')')


class TestGetch(TestCase):
    def setUp(self):
        self.patch('sys.stdin.fileno', self.mock_stdin_fileno)
        self.patch('select.select', self.mock_select)
        self.patch('os.read', self.mock_read)
        self.patch('tty.setraw', self.mock_setraw)
        self.patch('termios.tcgetattr', self.mock_tcgetattr)
        self.patch('termios.tcsetattr', self.mock_tcsetattr)
        self.buffer = bytearray()
        self.default_term_attr = [
                'iflag', 'oflag', 'cflag', 'lflag',
                'ispeed', 'ospeed',
                'cc']

        self.term_attr = list(self.default_term_attr)

    def tearDown(self):
        self.eq(self.term_attr, self.default_term_attr)

    def press(self, key):
        if isinstance(key, str):
            key = key.encode('utf8')
        self.buffer += key

    def mock_stdin_fileno(self):
        return 0

    def mock_select(self, rlist, wlist, xlist, timeout=None):
        self.eq(self.term_attr[0], 'raw')
        if self.buffer:
            return (rlist, [], [])
        return ([], [], [])

    def mock_read(self, fd, n):
        self.eq(self.term_attr[0], 'raw')
        ret = self.buffer[:n]
        del self.buffer[:n]
        return ret

    def mock_setraw(self, fd, when=None):
        import termios
        self.eq(when, termios.TCSADRAIN)
        self.term_attr = ['raw', 'raw', 'raw', 'raw', 'raw', 'raw', 'cc']

    def mock_tcgetattr(self, fd):
        return self.term_attr

    def mock_tcsetattr(self, fd, when, attributes):
        import termios
        self.eq(when, termios.TCSADRAIN)
        self.term_attr = attributes

    def test_getch_basic(self):
        self.eq(getch(), None)
        self.press(b'abc')
        self.eq(getch(), 'a')
        self.eq(getch(), 'b')
        self.eq(getch(), 'c')
        self.eq(getch(), None)

    def test_getch_unicode(self):
        self.eq(getch(), None)
        self.press('測試')
        self.eq(getch(), '測')
        self.eq(getch(), '試')
        self.eq(getch(), None)

    def test_getch_escape_keys(self):
        self.eq(getch(), None)
        self.press('\033[AA')
        self.eq(getch(), 'up')
        self.eq(getch(), 'A')
        self.eq(getch(), None)

    def test_getch_unicode_error(self):
        self.eq(getch(), None)
        test_data = '測'.encode('utf8')[:-1]
        self.press(test_data)
        self.eq(getch(), test_data)
        self.eq(getch(), None)

    def test_register_key_empty_seq(self):
        with self.raises(ValueError):
            register_key('')

    def test_register_key_with_key_object(self):
        new_key = type(KEY_UP)(r'\033[[[[[[', 'wow')
        nkey = register_key(new_key, 'wah', 'haha')
        self.eq(new_key.seq, nkey.seq)
        self.eq(new_key, 'wow')
        self.eq(nkey, 'wah')
        self.eq(nkey, 'haha')
        self.eq(deregister_key(new_key), new_key)

    def test_register_deregister_key(self):
        self.eq(getch(), None)
        self.press('測試')
        self.eq(getch(), '測')
        self.eq(getch(), '試')
        self.eq(getch(), None)

        # Resigter keys
        TE = register_key('測'.encode('utf8'), 'TE')
        ST = register_key('試'.encode('utf8'), 'ST')
        ABCD = register_key('\033ABCD', 'ABCD')
        self.eq(TE, '測')
        self.eq(TE, '測'.encode('utf8'))
        self.eq(ST, '試')
        self.eq(ST, '試'.encode('utf8'))
        self.eq(ABCD, 'ABCD')
        self.eq(ABCD, '\033ABCD')

        self.press('測試\033ABCD')
        self.eq(getch(), TE)
        self.eq(getch(), ST)
        self.eq(getch(), 'ABCD')
        self.eq(getch(), None)

        # Deresigter keys
        TE = deregister_key(TE)
        ST = deregister_key(ST.seq)
        ABCD = deregister_key('\033ABCD')

        self.press('測試\033ABCD')
        self.eq(getch(), TE)
        self.eq(getch(), ST)
        self.eq(getch(), '\33A')
        self.eq(getch(), 'B')
        self.eq(getch(), 'C')
        self.eq(getch(), 'D')
        self.eq(getch(), None)

        MY_HOME = register_key(KEY_HOME.seq, 'MY_HOME')
        self.eq(MY_HOME, KEY_HOME)
        self.press(KEY_HOME.seq)
        self.eq(getch(), MY_HOME)



import os
import shutil

from pathlib import Path

from .lib_test_utils import *

import warawara as wara


class TestSh(TestCase):
    def setUp(self):
        self.root = Path(__file__).parent
        self.cwd = self.root
        wara.sh.dir_stack = []
        self.patch('os.chdir', self.mock_chdir)
        self.patch('os.getcwd', self.mock_getcwd)

    def mock_getcwd(self):
        if isinstance(self.cwd, Exception):
            raise self.cwd

        if self.cwd == 'invalid':
            raise FileNotFoundError('invalid')

        return str(self.cwd)

    def mock_chdir(self, path):
        if path == 'invalid':
            raise FileNotFoundError('invalid')

        path = Path(path)
        if path.is_absolute():
            self.cwd = path
        else:
            self.cwd = self.cwd / path

    def test_cwd(self):
        self.eq(wara.cwd(), Path.cwd())
        self.eq(wara.cwd('tmp'), self.root / 'tmp')
        self.eq(wara.cwd('invalid'), False)
        self.eq(wara.cwd(0), False)
        self.eq(wara.cwd(True), False)

    def test_pushd_invalid_path(self):
        here = wara.cwd()

        p = wara.pushd('invalid')
        self.false(p)

        p = wara.pushd(3)
        self.false(p)

        self.eq(wara.cwd(), here)

    def test_pushd_invalid_cwd(self):
        self.cwd = FileNotFoundError('deleted')

        self.eq(wara.cwd(), False)

        p = wara.pushd('tmp')
        self.false(p)

    def test_popd_to_invalid_path(self):
        here = wara.cwd()

        wara.pushd('tmp')

        def mock_chdir(path):
            raise FileNotFoundError('deleted')
        self.patch('os.chdir', mock_chdir)

        self.false(wara.popd())

        self.eq(wara.cwd(), here / 'tmp')

    def test_dirs_clear(self):
        here = wara.cwd()
        wara.pushd('tmp')
        wara.pushd('tmp2')

        self.eq(wara.dirs(), [here, here / 'tmp', here / 'tmp' / 'tmp2'])

        wara.dirs(clear=True)

        self.eq(wara.dirs(), [here / 'tmp' / 'tmp2'])

    def test_popd_all(self):
        here = wara.cwd()
        wara.pushd('tmp')
        wara.pushd('tmp2')

        self.eq(wara.dirs(), [here, here / 'tmp', here / 'tmp' / 'tmp2'])

        wara.popd(all=True)

        self.eq(wara.dirs(), [here])

    def test_pushd_popd_dirs(self):
        here = wara.cwd()

        with wara.pushd('tmp'):
            self.eq(wara.cwd(), here / 'tmp')

            wara.pushd('a')
            self.eq(wara.cwd(), here / 'tmp' / 'a')
            self.eq(wara.dirs(),
                    [
                        here,
                        here / 'tmp',
                        here / 'tmp' / 'a',
                    ])

            wara.cwd(here / 'tmp' / 'b')
            self.eq(wara.cwd(), here / 'tmp' / 'b')

            self.eq(wara.popd(), True)

            self.eq(wara.cwd(), here / 'tmp')
            self.eq(wara.dirs(),
                    [
                        here,
                        here / 'tmp',
                    ])

        self.eq(wara.cwd(), here)
        self.eq(wara.dirs(), [here])

        self.eq(wara.popd(), False)

    def test_home(self):
        from os.path import expanduser
        self.eq(str(wara.home()), expanduser('~'))

    def test_shrinkuser(self):
        HOME = wara.home()
        self.eq(wara.shrinkuser(HOME), '~')
        self.eq(wara.shrinkuser(str(HOME) + '/'), '~/')

        self.eq(wara.shrinkuser(HOME / 'bana'), '~/bana')
        self.eq(wara.shrinkuser(HOME / 'bana/'), '~/bana')
        self.eq(wara.shrinkuser(str(HOME) + '/bana/'), '~/bana/')

        self.eq(wara.shrinkuser('bana/na/'), 'bana/na/')
        self.eq(wara.shrinkuser('bana/na'), 'bana/na')



import unittest
import threading


from .internal_utils import exporter
export, __all__ = exporter()


@export
class Checkpoint:
    def __init__(self, testcase):
        self.testcase = testcase
        self.checkpoint = threading.Event()

    def set(self):
        self.checkpoint.set()

    def clear(self):
        self.checkpoint.clear()

    def wait(self):
        self.checkpoint.wait()

    def is_set(self):
        return self.checkpoint.is_set()

    def check(self, is_set=True):
        self.testcase.eq(
                self.checkpoint.is_set(),
                is_set,
                'Checkpoint was' + (' ' if self.checkpoint.is_set() else ' not ') + 'set')

    def __bool__(self):
        return self.is_set()


@export
class TestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eq = self.assertEqual
        self.ne = self.assertNotEqual
        self.le = self.assertLessEqual
        self.lt = self.assertLess
        self.ge = self.assertGreaterEqual
        self.gt = self.assertGreater
        self.true = self.assertTrue
        self.false = self.assertFalse
        self.raises = self.assertRaises

    def checkpoint(self):
        return Checkpoint(self)

    class run_in_thread:
        def __init__(self, func, args=tuple(), kwargs=dict()):
            self.func = func
            self.args = args
            self.kwargs = kwargs
            self.thread = None

        def __enter__(self, *args):
            if self.thread is not None:
                raise RuntimeError('Thread objects cannot be reused')
            self.thread = threading.Thread(target=self.func, args=self.args, kwargs=self.kwargs)
            self.thread.daemon = True
            self.thread.start()

        def __exit__(self, exc_type, exc_value, traceback):
            self.thread.join()

    def patch(self, name, side_effect):
        patcher = unittest.mock.patch(name, side_effect=side_effect)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing


@export
class RunMocker:
    def __init__(self):
        self.rules = {}

    def register(self, cmd, callback=None, *, stdout=None, stderr=None, returncode=None):
        if not isinstance(cmd, str):
            raise ValueError('cmd must be a str')

        if callback is not None and not isinstance(callback, Exception) and not callable(callback):
            raise TypeError('callback should be an Exception or a callable')

        by_callback = callback
        by_output = (stdout, stderr, returncode)

        if by_callback is None and by_output == (None, None, None):
            raise ValueError('Meaningless behavior')

        if by_callback is not None and by_output != (None, None, None):
            raise ValueError('Ambiguous behavior')

        if cmd not in self.rules:
            self.rules[cmd] = []

        if by_callback:
            if isinstance(by_callback, Exception):
                behavior = by_callback
            else:
                def behavior(proc, *args):
                    proc.cmd = (cmd, *args)
                    return by_callback(proc, *args)

        else:
            def behavior(proc, *args):
                proc.cmd = (cmd, *args)
                if by_output[0]:
                    proc.stdout.writelines(by_output[0])
                if by_output[1]:
                    proc.stderr.writelines(by_output[1])
                return by_output[2]

        self.rules[cmd].append(behavior)
        return self

    def __call__(self, cmd, *,
                 stdin=None, stdout=True, stderr=True,
                 encoding='utf8', rstrip='\r\n',
                 bufsize=-1,
                 env=None,
                 wait=True):
        if not cmd:
            raise ValueError('command is empty')

        if isinstance(cmd, str):
            cmd = [cmd]

        matched_callbacks = None

        if cmd[0] in self.rules:
            matched_callbacks = self.rules[cmd[0]]
        elif '*' in self.rules:
            matched_callbacks = self.rules['*']
        else:
            raise ValueError('Unregistered command: {}'.format(cmd))

        behavior = matched_callbacks[0]
        if len(matched_callbacks) > 1:
            matched_callbacks.pop(0)

        if isinstance(behavior, Exception):
            raise behavior

        from .lib_subproc import command
        p = command([behavior] + cmd[1:],
                    stdin=stdin, stdout=stdout, stderr=stderr,
                    encoding=encoding, rstrip=rstrip,
                    bufsize=bufsize,
                    env=env)
        p.run(wait=wait)
        return p



import re
import abc
import itertools

from .lib_math import sgn
from .lib_math import vector
from .lib_math import lerp
from .lib_math import interval
from .lib_math import resample
from .lib_math import is_uint8
from .lib_math import clamp

from .internal_utils import exporter
export, __all__ = exporter()


@export
class Color(abc.ABC):
    @abc.abstractmethod
    def __init__(self, *args, **kwargs): # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self): # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def __int__(self): # pragma: no cover
        raise NotImplementedError

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.seq == other.seq

    def __call__(self, *args):
        return self.fg(*args)

    def fg(self, *args):
        return self.apply('38', ' '.join(str(arg) for arg in args))

    def bg(self, *args, **kwargs):
        return self.apply('48', ' '.join(str(arg) for arg in args))

    def apply(self, ground, s):
        if not self.seq:
            return s
        return '\033[{};{}m{}\033[m'.format(ground, self.seq, str(s))

    def __str__(self):
        return '\033[38;{}m'.format(self.seq) if self.seq else '\033[m'

    def __invert__(self):
        return ColorCompound(bg=self)

    def __truediv__(self, other):
        if not isinstance(other, Color):
            raise TypeError('Only Color() / Color() is allowed')
        return ColorCompound(fg=self, bg=other)

    def __or__(self, other):
        if isinstance(other, Color):
            return other if other.seq else self
        return ColorCompound(fg=self) | other


@export
def color(*args, **kwargs):
    nargs = len(args)
    arg1 = args[0] if len(args) == 1 else None

    # empty
    if not args:
        return Color256(None)

    if nargs != 1:
        # ColorRGB ctor
        if len(args) == 3 and all(is_uint8(i) for i in args):
            return ColorRGB(*args, **kwargs)

    # Copy ctor
    elif issubclass(type(arg1), Color):
        return type(arg1)(*args, **kwargs)

    # Color256 ctor
    elif arg1 is None or is_uint8(arg1):
        return Color256(*args, **kwargs)

    # ColorRGB ctor #RRGGBB
    elif isinstance(arg1, str) and re.fullmatch(r'#[0-9A-Fa-f]{6}', arg1):
        return ColorRGB(*args, **kwargs)

    # ColorHSV @H,S,V format
    elif isinstance(arg1, str) and re.fullmatch(r'@[0-9]+,[0-9]+,[0-9]+', arg1):
        return ColorHSV(*args, **kwargs)

    raise TypeError('Invalid arguments: {}'.format(args))


@export
class Color256(Color):
    def __init__(self, index=None):
        if isinstance(index, self.__class__):
            index = index.index

        self.index = index

        if not self.index is None and not is_uint8(index):
            raise TypeError('Invalid color index: {}'.format(index))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.index)

    @property
    def code(self):
        return self.index

    @property
    def seq(self):
        if self.index is None:
            return ''
        return '5;{}'.format(self.index)

    def to_rgb(self):
        if self.index < 16:
            base = 0xFF if (self.index > 7) else 0x80
            is_7 = (self.index == 7)
            is_8 = (self.index == 8)
            R = base * ((self.index & 0x1) != 0) + (0x40 * is_7) + (0x80 * is_8)
            G = base * ((self.index & 0x2) != 0) + (0x40 * is_7) + (0x80 * is_8)
            B = base * ((self.index & 0x4) != 0) + (0x40 * is_7) + (0x80 * is_8)

        elif self.index < 232:
            base = self.index - 16
            index_R = (base // 36)
            index_G = ((base % 36) // 6)
            index_B = (base % 6)
            R = (55 + index_R * 40) if index_R > 0 else 0
            G = (55 + index_G * 40) if index_G > 0 else 0
            B = (55 + index_B * 40) if index_B > 0 else 0

        else:
            R = G = B = (self.index - 232) * 10 + 8

        return ColorRGB(R, G, B)

    def to_hsv(self):
        return self.to_rgb().to_hsv()

    def __int__(self):
        return self.index


@export
class ColorRGB(Color):
    def __init__(self, *args, overflow=False):
        nargs = len(args)
        arg1 = args[0] if len(args) else None

        type_check = (lambda x: 0 <= x < 256
                      if not overflow
                      else lambda x: isinstance(x, (int, float)))

        self.r = None
        self.g = None
        self.b = None

        if not args:
            return

        elif nargs != 1:
            # (num, num, num) format
            if len(args) == 3 and all(type_check(i) for i in args):
                (self.r, self.g, self.b) = args

        # Copy ctor
        elif isinstance(arg1, self.__class__):
            (self.r, self.g, self.b) = arg1.rgb

        # #RRGGBB format
        elif isinstance(arg1, str) and re.fullmatch(r'#[0-9A-Fa-f]{6}', arg1):
            rgb_str = arg1[1:]
            self.r = int(rgb_str[0:2], 16)
            self.g = int(rgb_str[2:4], 16)
            self.b = int(rgb_str[4:6], 16)

        if None in self.rgb:
            raise TypeError('Invalid RGB value: {}'.format(args))

    def __repr__(self):
        return 'ColorRGB({}, {}, {})'.format(self.r, self.g, self.b)

    @property
    def R(self):
        return 0 if self.r is None else clamp(0, round(self.r), 255)

    @property
    def G(self):
        return 0 if self.g is None else clamp(0, round(self.g), 255)

    @property
    def B(self):
        return 0 if self.b is None else clamp(0, round(self.b), 255)

    @property
    def RGB(self):
        return (self.R, self.G, self.B)

    @property
    def rgb(self):
        return (self.r, self.g, self.b)

    @property
    def seq(self):
        if None in self.rgb:
            return ''
        return '2;{};{};{}'.format(self.R, self.G, self.B)

    def __add__(self, other):
        rgb = vector(self.rgb) + vector(other.rgb)
        return ColorRGB(*rgb, overflow=True)

    def __mul__(self, num):
        return ColorRGB(*vector(self.rgb) * num, overflow=True)

    def __floordiv__(self, num):
        return ColorRGB(*vector(self.rgb) // num, overflow=True)

    def __int__(self):
        return (self.R << 16) | (self.G << 8) | (self.B)

    def __format__(self, spec):
        if not spec:
            return str(self)

        if spec in ('#', '#x', '#X'):
            x = (spec + 'X')[1]
            return '#{r:0>2{x}}{g:0>2{x}}{b:0>2{x}}'.format(r=self.R, g=self.G, b=self.B, x=x)

        return format(self.RGB, spec)

    def to_rgb(self):
        return self

    def to_hsv(self, overflow=False):
        import colorsys
        hsv = colorsys.rgb_to_hsv(self.R / 255, self.G / 255, self.B / 255)
        return ColorHSV(hsv[0] * 360, hsv[1] * 100, hsv[2] * 100, overflow=overflow)


@export
class ColorHSV(Color):
    def __init__(self, *args, overflow=False):
        arg1 = args[0] if len(args) else None

        self.h = None
        self.s = None
        self.v = None

        h = None
        s = None
        v = None

        if not args:
            return

        # Copy ctor
        elif len(args) == 1 and isinstance(arg1, self.__class__):
            other = arg1
            (h, s, v) = (other.h, other.s, other.v)

        # @H,S,V format
        elif len(args) == 1 and isinstance(arg1, str) and re.fullmatch(r'@[0-9]+,[0-9]+,[0-9]+', arg1):
            (h, s, v) = list(map(lambda x: int(x, 10), arg1[1:].split(',')))

        # (num, num, num) format
        elif len(args) == 3:
            (h, s, v) = args

        # Value range check
        if (all(isinstance(x, (int, float)) for x in (h, s, v)) and
            overflow or
            ((s is not None and 0 <= s <= 100) and
             (v is not None and 0 <= v <= 100))):
            (self.h, self.s, self.v) = (h % 360, s, v)
            self._rgb = self.to_rgb(overflow)

        else:
            raise TypeError('Invalid HSV value: {}'.format(args))

    def __repr__(self):
        return 'ColorHSV({:}deg, {:}%, {:}%)'.format(*self.HSV)

    @property
    def H(self):
        return 0 if self.h is None else ((round(self.h) + 360) % 360)

    @property
    def S(self):
        return 0 if self.s is None else clamp(0, round(self.s), 100)

    @property
    def V(self):
        return 0 if self.v is None else clamp(0, round(self.v), 100)

    @property
    def HSV(self):
        return (self.H, self.S, self.V)

    @property
    def hsv(self):
        return (self.h, self.s, self.v)

    @property
    def seq(self):
        if None in self.hsv:
            return ''
        return self._rgb.seq

    def __add__(self, other):
        hsv = vector(self.hsv) + vector(other.hsv)
        return ColorHSV(*hsv, overflow=True)

    def __mul__(self, num):
        return ColorHSV(*vector(self.hsv) * num, overflow=True)

    def __floordiv__(self, num):
        return ColorHSV(*vector(self.hsv) // num, overflow=True)

    def __int__(self):
        return (self.H * 1000000) + (self.S * 1000) + (self.V)

    def __format__(self, spec):
        if not spec:
            return str(self)
        if spec == '#':
            return '(@{}, {}%, {}%)'.format(self.H, self.S, self.V)
        return format(self.hsv, spec)

    def to_rgb(self, overflow=False):
        import colorsys
        return ColorRGB(*vector(colorsys.hsv_to_rgb(
            self.H / 360,
            self.S / 100,
            self.V / 100)) * 255, overflow=overflow)

    def to_hsv(self):
        return self


@export
class ColorCompound:
    def __init__(self, fg=None, bg=None):
        self.fg = color(fg)
        self.bg = color(bg)

        seq = ';'.join(filter(None, [
            '38;' + self.fg.seq if self.fg.seq else None,
            '48;' + self.bg.seq if self.bg.seq else None,
            ]))
        self.seq = '' if not seq else ('\033[' + seq + 'm')

    def __repr__(self):
        return 'ColorCompound(fg={fg}, bg={bg})'.format(fg=self.fg, bg=self.bg)

    def __call__(self, s=''):
        return s if not self.seq else f'{self.seq}{s}\033[m'

    def __str__(self):
        return self.seq or '\033[m'

    def __or__(self, other):
        fg = other.fg if other.fg.seq else self.fg
        bg = other.bg if other.bg.seq else self.bg

        return ColorCompound(fg=fg, bg=bg)

    def __truediv__(self, other):
        return ColorCompound(fg=self.fg, bg=other.fg)

    def __invert__(self):
        return ColorCompound(fg=self.bg, bg=self.fg)

    def __eq__(self, other):
        return self.seq == other.seq

@export
def paint(fg=None, bg=None):
    return ColorCompound(fg=fg, bg=bg)


export('nocolor')
nocolor = color()

named_colors = [
        (0, ('black',)),
        (1, ('maroon',)),
        (2, ('green',)),
        (3, ('olive',)),
        (4, ('navy',)),
        (5, ('purple',)),
        (6, ('teal',)),
        (7, ('silver',)),
        (8, ('gray', 'grey',)),
        (9, ('red',)),
        (10, ('lime',)),
        (11, ('yellow',)),
        (12, ('blue',)),
        (13, ('fuchsia', 'magenta',)),
        (14, ('aqua', 'cyan',)),
        (15, ('aliceblue', 'azure', 'floralwhite', 'ghostwhite', 'ivory',
              'lavenderblush', 'mintcream', 'snow', 'white',)),
        (17, ('midnightblue',)),
        (18, ('darkblue',)),
        (20, ('mediumblue',)),
        (22, ('darkgreen',)),
        (28, ('forestgreen',)),
        (29, ('seagreen',)),
        (30, ('darkcyan',)),
        (33, ('dodgerblue',)),
        (37, ('lightseagreen',)),
        (39, ('deepskyblue',)),
        (44, ('darkturquoise',)),
        (48, ('mediumspringgreen', 'springgreen',)),
        (54, ('indigo',)),
        (60, ('darkslateblue',)),
        (62, ('royalblue', 'slateblue',)),
        (64, ('olivedrab',)),
        (66, ('slategray', 'slategrey',)),
        (67, ('steelblue',)),
        (69, ('cornflowerblue',)),
        (71, ('mediumseagreen',)),
        (73, ('cadetblue',)),
        (77, ('limegreen',)),
        (79, ('mediumaquamarine',)),
        (80, ('mediumturquoise', 'turquoise',)),
        (88, ('darkred',)),
        (90, ('darkmagenta',)),
        (92, ('blueviolet', 'darkviolet',)),
        (94, ('saddlebrown',)),
        (98, ('darkorchid', 'mediumpurple',)),
        (99, ('mediumslateblue',)),
        (102, ('lightslategray', 'lightslategrey',)),
        (108, ('darkseagreen',)),
        (113, ('yellowgreen',)),
        (117, ('lightskyblue', 'skyblue',)),
        (118, ('chartreuse', 'lawngreen',)),
        (120, ('lightgreen', 'palegreen',)),
        (122, ('aquamarine',)),
        (124, ('brown', 'firebrick',)),
        (130, ('sienna',)),
        (134, ('mediumorchid',)),
        (135, ('murasaki',)),
        (136, ('darkgoldenrod',)),
        (138, ('rosybrown',)),
        (143, ('darkkhaki',)),
        (152, ('lightblue', 'lightsteelblue', 'powderblue',)),
        (154, ('greenyellow',)),
        (159, ('paleturquoise',)),
        (161, ('crimson',)),
        (162, ('mediumvioletred',)),
        (166, ('chocolate', 'clementine',)),
        (167, ('indianred',)),
        (168, ('palevioletred',)),
        (170, ('orchid',)),
        (173, ('peru',)),
        (174, ('darksalmon',)),
        (178, ('goldenrod',)),
        (180, ('burlywood', 'tan',)),
        (182, ('plum', 'thistle',)),
        (195, ('lightcyan',)),
        (198, ('deeppink',)),
        (202, ('orangered',)),
        (203, ('tomato',)),
        (205, ('hotpink',)),
        (208, ('darkorange',)),
        (209, ('coral', 'salmon',)),
        (210, ('lightcoral',)),
        (213, ('violet',)),
        (214, ('orange',)),
        (215, ('sandybrown',)),
        (216, ('lightsalmon',)),
        (217, ('lightpink',)),
        (218, ('pink',)),
        (220, ('gold',)),
        (222, ('khaki',)),
        (223, ('moccasin', 'navajowhite', 'palegoldenrod',
               'peachpuff', 'wheat',)),
        (224, ('bisque', 'mistyrose',)),
        (230, ('antiquewhite', 'beige', 'blanchedalmond',
               'cornsilk', 'lemonchiffon',
               'lightgoldenrodyellow', 'lightyellow',
               'oldlace', 'papayawhip',)),
        (238, ('darkslategray', 'darkslategrey',)),
        (239, ('darkolivegreen',)),
        (242, ('dimgray', 'dimgrey',)),
        (248, ('darkgray', 'darkgrey',)),
        (252, ('lightgray', 'lightgrey',)),
        (253, ('gainsboro',)),
        (255, ('honeydew', 'lavender', 'linen', 'seashell', 'whitesmoke',)),
]
export('names')
names = tuple(name for index, names in named_colors for name in names)
def _setup_named_colors():
    for index, names in named_colors:
        clr = color(index)
        for name in names:
            globals()[name] = clr
            export(name)
_setup_named_colors()
del _setup_named_colors


decolor_regex = re.compile('\033' + r'\[[\d;]*m')

@export
def decolor(s):
    return decolor_regex.sub('', s)


@export
def gradient(A, B, N=None, reverse=False, clockwise=None):
    if not isinstance(A, Color) or not isinstance(B, Color):
        raise TypeError('Can only calculate gradient() on Color objects')

    if N is not None and not isinstance(N, int):
        raise TypeError('N must be a integer')

    if N is not None and N < 2:
        raise ValueError('N={} is too small'.format(N))

    ret = None
    if N == 2:
        ret = (A, B)

    elif isinstance(A, Color256) and isinstance(B, Color256):
        ret = gradient_color256(A, B, N=N)

    elif isinstance(A, ColorRGB) and isinstance(B, ColorRGB):
        ret = gradient_rgb(A, B, N=N)

    elif isinstance(A, ColorHSV) and isinstance(B, ColorHSV):
        ret = gradient_hsv(A, B, N=N, clockwise=clockwise)

    else:
        ret = (A, B)

    if reverse:
        return ret[::-1]
    else:
        return ret


def gradient_color256(A, B, N=None):
    if A.index in range(232, 256) and B.index in range(232, 256):
        return gradient_color256_grayscale_range(A, B, N)

    if A.index in range(16, 232) and B.index in range(16, 232):
        return gradient_color256_rgb_range(A, B, N)

    return (A, B)


def gradient_color256_grayscale_range(A, B, N=None):
    a, b = A.index, B.index
    direction = sgn(b - a)
    n = abs(b - a) + 1
    return tuple(Color256(c) for c in resample(interval(a, b), N or n))


def gradient_color256_rgb_range(A, B, N=None):
    def color_to_rgb6(p):
        c = int(p) - 16
        r = c // 36
        g = (c % 36) // 6
        b = c % 6
        return vector(r, g, b)

    def rgb6_to_color(rgb6):
        return Color256(rgb6[0] * 36 + rgb6[1] * 6 + rgb6[2] + 16)

    rgb_a = color_to_rgb6(A)
    rgb_b = color_to_rgb6(B)

    delta = rgb_b - rgb_a
    cont_step_count = max(abs(d) for d in delta)

    if N is None or N > cont_step_count:
        # N >= minimum contiguous path
        steps = []
        for n in range(cont_step_count):
            step = delta.map(sgn)
            steps.append(step)
            delta = delta.map(lambda x: x - sgn(x))

        ret = resample(list(itertools.accumulate([rgb_a] + steps)), N)

    else:
        # N is shorter than minimum contiguous path
        ret = zip(
                resample(interval(rgb_a[0], rgb_b[0]), N),
                resample(interval(rgb_a[1], rgb_b[1]), N),
                resample(interval(rgb_a[2], rgb_b[2]), N),
                )

    return tuple(rgb6_to_color(i) for i in ret)


def gradient_rgb(A, B, N):
    # Calculate gradient in RGB
    a = vector(A.rgb)
    b = vector(B.rgb)
    if N is None:
        import math
        dist_r = math.ceil(abs(a[0] - b[0]) // 40)
        dist_g = math.ceil(abs(a[1] - b[1]) // 40)
        dist_b = math.ceil(abs(a[2] - b[2]) // 40)
        N = max(dist_r, dist_g, dist_b)

    ret = [A]
    for t in (i / (N - 1) for i in range(1, N - 1)):
        ret.append(ColorRGB(*tuple(lerp(a, b, t))))
    ret.append(B)
    return tuple(ret)


def gradient_hsv(A, B, N, clockwise):
    # Calculate gradient in HSV
    a = vector(A.hsv)
    b = vector(B.hsv)

    if clockwise == True:
        b[0] += 360 if (a[0] > b[0]) else 0
    elif clockwise == False:
        a[0] += 360 if (a[0] < b[0]) else 0

    if N is None:
        import math
        dist_hue = math.ceil(abs(a[0] - b[0]) / 30)
        dist_sat = math.floor(abs(a[1] - b[1]) / 10)
        dist_val = math.floor(abs(a[2] - b[2]) / 10)
        N = max(dist_hue, dist_sat, dist_val)

    ret = [A]
    for t in (i / (N - 1) for i in range(1, N - 1)):
        c = lerp(a, b, t)
        ret.append(ColorHSV(*c))

    ret.append(B)

    return tuple(ret)



import builtins

from .internal_utils import exporter
export, __all__ = exporter()


class LineFileWrapper:
    def __init__(self, path, mode, rstrip, newline, **kwargs):
        self.path = path
        self.mode = mode
        self.rstrip = rstrip
        self.newline = newline
        self.kwargs = kwargs

        self.file = builtins.open(self.path, mode=self.mode, **self.kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        cleanup = getattr(self.file, '__exit__')
        if cleanup:
            return cleanup(*args, **kwargs)

        else: # pragma: no cover
            cleanup = getattr(self.file, 'close')
            if cleanup:
                cleanup()

    def __getattr__(self, attr):
        return getattr(self.file, attr)

    def writeline(self, *args):
        self.file.write(' '.join(str(arg) for arg in args) + self.newline)

    def writelines(self, lines):
        for line in lines:
            self.writeline(line)

    def readline(self):
        return self.file.readline().rstrip(self.rstrip)

    def readlines(self):
        return [line for line in self]

    def __iter__(self):
        for line in self.file:
            yield line.rstrip(self.rstrip)


@export
def open(path, mode='rt', rstrip='\r\n', newline='\n', **kwargs):
    # Skip for binary mode
    if 'b' in mode:
        return builtins.open(path, mode=mode, **kwargs)

    kwargs['encoding'] = kwargs.get('encoding', 'utf-8')
    kwargs['errors'] = kwargs.get('errors', 'backslashreplace')

    return LineFileWrapper(path, mode, rstrip=rstrip, newline=newline, **kwargs)


@export
def natsorted(iterable, key=None):
    import re
    def filename_as_key(name):
        def int_or_not(x):
            if x and x[0] in '1234567890':
                return int(x)
            return x
        return tuple(int_or_not(x) for x in re.split(r'([0-9]+)', str(name)))

    sort_key = (filename_as_key
                if key is None
                else lambda x: filename_as_key(key(x))
                )

    return sorted(iterable, key=sort_key)



from .lib_test_utils import *

from .internal_utils import exporter


class TestInternalUtils(TestCase):
    def test_exporter(self):
        export, uualluu = exporter()

        self.eq(uualluu, [])

        export('wara')
        self.eq(uualluu, ['wara'])

        @export
        def test_func():
            pass

        self.eq(uualluu, ['wara', 'test_func'])

        export()
        self.eq(uualluu, ['wara', 'test_func'])



def exporter():
    # Inspired by https://stackoverflow.com/a/41895194/3812388

    uualluu = []

    def decorator(*args):
        if not args:
            return

        for what in args:
            if isinstance(what, str):
                uualluu.append(what)
            else:
                uualluu.append(what.__name__)

        return what

    return decorator, uualluu


export, __all__ = exporter()
export(exporter)



import re

from .internal_utils import exporter
export, __all__ = exporter()


@export
class rere:
    def __init__(self, text):
        self.text = text
        self.cache = None

    def search(self, pattern, flags=0):
        self.cache = re.search(pattern, self.text, flags=flags)
        return self.cache

    def match(self, pattern, flags=0):
        self.cache = re.match(pattern, self.text, flags=flags)
        return self.cache

    def fullmatch(self, pattern, flags=0):
        self.cache = re.fullmatch(pattern, self.text, flags=flags)
        return self.cache

    def sub(self, pattern, repl):
        return re.sub(pattern, repl, self.text)

    def __getattr__(self, attr):
        if hasattr(re, attr):
            re_attr = getattr(re, attr)
            return lambda *args, **kwargs: re_attr(*args, self.text, **kwargs)

        return getattr(self.cache, attr)
